from __future__ import print_function, division, absolute_import

from collections import Mapping

import tensorflow as tf

from odin.utils import (as_tuple, is_string, is_number,
                        wprint, one_hot, ctext)

from odin.training.trainer import Timer, MainLoop
from odin.training.callbacks import *

# ===========================================================================
# Helper
# ===========================================================================
def _parse_optimizer(name):
  from odin import backend as K
  name = str(name).lower()
  for key in dir(K.optimizers):
    obj = getattr(K.optimizers, key)
    if isinstance(obj, type) and issubclass(obj, K.optimizers.Optimizer):
      # more robust in handling the name
      if name in key.lower():
        return obj
  return None

def _preprocessing_data(train, valid):
  from odin import fuel as F
  train = F.as_data(train)
  if is_number(valid):
    start_train = 0.
    end_train = 1. - valid
    start_valid = 1. - valid
    end_valid = 1.
    valid = F.DataGroup(train.data).set_batch(start=start_valid, end=end_valid)
    train = F.DataGroup(train.data).set_batch(start=start_train, end=end_train)
  elif valid is not None:
    valid = F.as_data(valid)
  return train, valid

def _preprocessing_losses(losses, y_true, y_pred, inherit_losses=None,
                          sample_weights=None):
  """ Can be used for both objectives and metrics """
  from odin import backend as K
  # ====== special cases, only one inputs outputs, and multiple loss ====== #
  nb_losses = len(losses)
  if len(y_true) == 0:
    y_true = [None] * nb_losses
  elif len(y_true) == 1:
    y_true = y_true * nb_losses
  if len(y_pred) == 0:
    y_pred = [None] * nb_losses
  elif len(y_pred) == 1:
    y_pred = y_pred * nb_losses
  # ====== applying ====== #
  cost = []
  for idx, fn in enumerate(as_tuple(losses)):
    weight = 1
    kwargs = {}
    # preprocess
    if isinstance(fn, (tuple, list)):
      if len(fn) == 1:
        fn = fn[0]
      else:
        weight = [i for i in fn if is_number(i)]
        weight = 1 if len(weight) == 0 else weight[0]
        kwargs = [i for i in fn if isinstance(i, Mapping)]
        kwargs = {} if len(kwargs) == 0 else kwargs[0]
        fn = [i for i in fn if i != weight and i != kwargs][0]
    # apply the loss
    if is_number(fn):
      if inherit_losses is None or fn >= len(inherit_losses):
        raise ValueError("Cannot find losses at index: '%d'" % fn)
      obj = inherit_losses[fn]
    elif K.is_tensor(fn):
      obj = fn
    elif hasattr(fn, '__call__'):
      try:
        sign = inspect.signature(fn)
        if 'weights' in sign.parameters and sample_weights is not None:
          kwargs['weights'] = sample_weights
      except ValueError:
        pass
      finally:
        obj = fn(y_true[idx], y_pred[idx], **kwargs)
      if isinstance(obj, (tuple, list)):
        wprint("function: '%s' return %d outputs (%s), only pick the first one"
               % (fn.__name__,
                  len(obj),
                  '; '.join([str(i) for i in obj])))
        obj = obj[0]
    cost.append((weight, obj))
  # ====== reduce ====== #
  return [c if w == 1 else w * c for w, c in cost]

def _preprocess_prior_weights(y_true, prior_weights):
  if prior_weights is None:
    return None
  from odin import backend as K
  # ====== everything must be list ====== #
  if not isinstance(prior_weights, (tuple, list)):
    prior_weights = (prior_weights,)
  elif is_number(prior_weights[0]):
    prior_weights = (prior_weights,)
  # ====== matching indices and prior_weights ====== #
  pw = 0
  for yt, w in zip(y_true, prior_weights):
    if w is not None:
      pw += K.to_sample_weights(indices=yt, weights=w)
  return pw

# ===========================================================================
# Main methods
# ===========================================================================
def train(X, y_true, y_pred, train_data,
          valid_data=None, valid_freq=1.,
          patience=3, threshold=5, rollback=True,
          objectives=[tf.losses.softmax_cross_entropy],
          metrics=[0], training_metrics=[],
          l1_regu=0., l2_regu=0., parameters=[],
          prior_weights=None, sample_weights=None,
          batch_size=256, epochs=8, shuffle=True,
          optimizer='rmsprop', optz_kwargs={'lr': 0.001}, updates=None,
          init_vars=True, labels=None, seed=5218, verbose=2):
  """

  Parameters
  ----------
  rollback : bool (default: True)
    if True, allow rollback to the best checkpoint during training
  objectives : {callable, tensorflow.Tensor}
    if `callable`, the function must take `y_true`, and `y_pred`
    The objectives must be differentiable and used for training.
  metrics : {callable, tensorflow.Tensor, int}
    if `callable`, the function must take `y_true`, and `y_pred`
    The `metrics` is for monitoring the training process.
    if `int`, it is the index of the loss in `objectives`
    NOTE: the first metrics in the list will be used for
    early-stopping (smaller is better).
  training_metrics : {callable, tensorflow.Tensor, int}
    if `int`, it is the index of the loss in `metrics`
  parameters : {list or tensorflow.Variables}
    All the parameters will be updated by the `optimizer`, if None
    or empty list is given, use ComputationalGraph to get
    all variables with Parameters roles related to the objectives
  init_vars : bool (default: True)
    automatically initialize all variables
  labels : {None, list of string}
    Given labels for classification task
  seed : int
    specific random seed for reproducible
  verbose : int
    0 - Turn off all log
    1 - only show notification
    2 - show notification, important log and summary
    3 - Show progress, summary, notification and logging
    4 - Show debug information and everything

  Return
  ------
  Function used for prediction
  """
  from odin import backend as K
  # ====== preprocess inputs ====== #
  X = as_tuple(X, t=K.is_tensor)
  y_true = as_tuple(y_true, t=K.is_tensor)
  y_pred = as_tuple(y_pred, t=K.is_tensor)
  # ====== parsing objectives and metrics ====== #
  # for training
  prior_weights = _preprocess_prior_weights(y_true=y_true,
                                            prior_weights=prior_weights)
  if prior_weights is not None:
    if sample_weights is not None:
      sample_weights = sample_weights + prior_weights
    else:
      sample_weights = prior_weights
  objectives = _preprocessing_losses(as_tuple(objectives), y_true, y_pred,
                                     sample_weights=sample_weights)
  # metrics for monitoring
  metrics = as_tuple(metrics)
  get_value = lambda x: np.mean(x)
  if len(metrics) > 0 and \
  (metrics[0] == tf.metrics.accuracy or
   metrics[0] == K.metrics.categorical_accuracy):
    get_value = lambda x: 1 - np.mean(x)
  metrics = _preprocessing_losses(metrics, y_true, y_pred,
                                  inherit_losses=objectives)
  # training_metrics
  training_metrics = _preprocessing_losses(as_tuple(training_metrics),
                                           y_true, y_pred,
                                           inherit_losses=metrics)
  # sum the objectives for differentiable
  if len(objectives) > 0:
    objectives = [sum(objectives) if len(objectives) > 1 else objectives[0]]
  # ====== preprocess optimizer and get updates====== #
  if updates is None: # not given updates
    if is_string(optimizer):
      optimizer = _parse_optimizer(optimizer)
      optimizer = optimizer(**optz_kwargs)
    elif not isinstance(optimizer, K.optimizers.Optimizer):
      raise ValueError("`optimizer` must be string - name of algorithm or instance "
                       "of odin.backend.optimizers.Optimizer")
    parameters = K.ComputationGraph(objectives).parameters\
    if len(parameters) == 0 else as_tuple(parameters, t=K.is_variable)
    # check objectives
    if len(objectives) == 0:
      raise RuntimeError("`objectives` must be given due to `updates=None`")
    weights = [p for p in parameters if K.role.has_roles(p, roles=K.role.Weight)]
    # l1 regularization
    if l1_regu > 0.:
      l1_norm = sum(tf.norm(w, ord=1) for w in weights)
      objectives[0] += l1_norm
    # l2 regularization
    if l2_regu > 0.:
      l2_norm = sum(tf.norm(w, ord=2) for w in weights)
      objectives[0] += l2_norm
    # update rules
    updates = optimizer.get_updates(objectives[0], parameters)
    # adding global norm and learning rate
    training_metrics.append(optimizer.norm)
    training_metrics.append(optimizer.lr)
  elif K.is_operation(updates): # given updates
    optimizer = None
  else:
    raise ValueError("`updates` can be None or tensorflow Operation, but given "
      "type: %s" % str(type(updates)))
  # ====== placeholders ====== #
  inputs_plh = []
  for plh in X:
    for i in (K.ComputationGraph(plh).placeholders
              if not K.is_placeholder(plh)
              else as_tuple(plh)):
      inputs_plh.append(i)
  outputs_plh = []
  for plh in y_true: # no duplicated inputs (e.g. autoencoder X == y)
    if not K.is_placeholder(plh):
      plh = K.ComputationGraph(plh).placeholders
    for i in as_tuple(plh):
      if i not in inputs_plh:
        outputs_plh.append(i)
  inputs = inputs_plh + outputs_plh
  # ====== initialize variables ====== #
  if bool(init_vars):
    K.initialize_all_variables()
  # ====== creating function ====== #
  # training function
  f_train = K.function(inputs=inputs,
                       outputs=objectives + training_metrics,
                       updates=updates, training=True)
  # scoring function
  f_score = None
  if len(metrics) > 0:
    f_score = K.function(inputs=inputs, outputs=metrics,
                         training=False)
  # prediction function
  f_pred = K.function(inputs=inputs_plh,
                      outputs=y_pred[0] if len(y_pred) == 1 else y_pred,
                      training=False)
  # ====== preprocessing data ====== #
  train_data, valid_data = _preprocessing_data(train_data, valid_data)
  # print some debug information if necessary
  if verbose >= 4:
    print("%s %s %s" % (
        ctext("============", 'cyan'),
        ctext("Prepare for Training", 'red'),
        ctext("============", 'cyan')))
    print(ctext("Input placeholders:", 'yellow'))
    for i in inputs_plh:
      print(" * ", str(i))
    print(ctext("Output placeholders:", 'yellow'))
    for i in outputs_plh:
      print(" * ", str(i))
    print(ctext("Parameters:", 'yellow'))
    for p in parameters:
      print(" * ", p.name, '-', p.shape, ';', p.dtype.name)
    print(ctext("Optimizer:", 'yellow'))
    print(" * ", str(optimizer))
    print(" * Optimizer kwargs:", optz_kwargs)
    print(" * L1:", l1_regu)
    print(" * L2:", l2_regu)
    print(ctext("Training:", 'yellow'))
    print(" * Valid freq:", valid_freq)
    print(" * Patience:", patience)
    print(" * Threshold:", threshold)
    print(" * Rollback:", rollback)
    print(" * Batch size:", batch_size)
    print(" * Epoch:", epochs)
    print(" * Shuffle:", shuffle)
    print(" * Seed:", seed)
    print(ctext("Objectives:", 'yellow'))
    for o in objectives:
      print(" * ", str(o))
    print(ctext("Weights:", 'yellow'))
    print(" * Prior:", str(prior_weights))
    print(" * Sample:", str(sample_weights))
    print(ctext("Metrics:", 'yellow'))
    for m in metrics:
      print(" * ", str(m))
    print(ctext("Training metrics:", 'yellow'))
    for t in training_metrics:
      print(" * ", str(t))
    print(ctext("Training Data:", 'yellow'), str(train_data))
    print(ctext("Validating Data:", 'yellow'), str(valid_data))
    print(ctext("Labels:", 'yellow'), labels)
  # ====== create trainer ====== #
  callback_log = True if verbose > 0 else False
  trainer = MainLoop(batch_size=batch_size,
                     seed=seed if shuffle else None,
                     shuffle_level=2 if shuffle else 0,
                     allow_rollback=rollback,
                     verbose=verbose, labels=labels)
  trainer.set_checkpoint(path=None, obj=None,
                         variables=parameters)
  # create callback
  callbacks = [NaNDetector(patience=patience, log=callback_log)]
  if valid_data is not None and f_score is not None:
    callbacks.append(
        EarlyStopGeneralizationLoss(task_name='valid', output_name=metrics[0],
                                    threshold=threshold, patience=patience,
                                    log=callback_log, get_value=get_value))
  trainer.set_callbacks(callbacks)
  # set the tasks
  trainer.set_train_task(func=f_train, data=train_data,
                         epoch=epochs, name='train')
  if valid_data is not None and f_score is not None:
    trainer.set_valid_task(func=f_score, data=valid_data,
                           freq=Timer(percentage=valid_freq),
                           name='valid')
  # running
  trainer.run()
  return f_pred



from tensorflow.python.util import nest
from tensorflow.python.keras.layers import Layer

__all__ = [
    'to_keras_objective',
    'copy_keras_metadata',
    'has_keras_meta',
    'add_trainable_weights'
]

def tied_session():
  """ Tied the tensorflow Session and keras Session together
  """
  from odin.autoconfig import get_session
  from tensorflow.python.keras.backend import set_session
  set_session(get_session())

def to_keras_objective(tensor, name=None):
  """ Convert any tensor to an objective function
  (i.e. loss or metric function) for keras fit Model
  """
  if name is None:
    name = tensor.name
  name = name.strip()
  if ':' in name:
    name = name.split(':')[0]
  fn = lambda *args: tensor
  fn.__name__ = str(name)
  return fn

def has_keras_meta(tensor):
  return hasattr(tensor, '_keras_history') and hasattr(tensor, '_keras_mask')

def copy_keras_metadata(keras_tensor, *new_tensors):
  if not hasattr(keras_tensor, '_keras_history') or \
  not hasattr(keras_tensor, '_keras_mask'):
    pass
  else:
    new_tensors = nest.flatten(new_tensors)
    history = keras_tensor._keras_history
    mask = keras_tensor._keras_mask
    for t in new_tensors:
      setattr(t, '_keras_history', history)
      setattr(t, '_keras_mask', mask)
  return new_tensors[0] if len(new_tensors) == 1 else new_tensors

def add_trainable_weights(layer, *variables):
  from odin.backend import is_variable
  variables = nest.flatten(variables)
  assert all(is_variable(v) for v in variables), \
  "All objects from variables must be instance of tensorflow.Variable"

  assert isinstance(layer, Layer), \
  "layer must be instance of tensorflow.python.keras.layers.Layer"

  variables = [v for v in variables
               if v not in layer._trainable_weights]
  layer._trainable_weights = layer._trainable_weights + variables
  return layer



"""Tests for the ZeroInflated distribution."""
from __future__ import absolute_import, division, print_function

# Dependency imports
import numpy as np

import tensorflow as tf
import tensorflow_probability as tfp

from tensorflow_probability.python.internal import test_util as tfp_test_util
from tensorflow.python.framework import test_util  # pylint: disable=g-direct-tensorflow-import
tfd = tfp.distributions

from odin.bay import ZeroInflated

def make_synthetic_data(shape, rate,
      count_distribution=lambda size: np.random.negative_binomial(n=5, p=0.3, size=size)):
  """ Generating samples according to Zero-inflation process
  shape : output shape
  rate : zero-inflation rate
  count_distribution : distribution

  Return
  ------
  data_corrupted : corrupted data by zero inflation
  data : original data
  mask : indicator which entries were dropped out
  """
  data = count_distribution(size=shape)
  mask = np.random.binomial(n=1, p=np.clip(1. - rate, 0., 1.),
                            size=shape)
  data_corrupted = (data * mask)
  return data_corrupted, data, 1 - mask

@test_util.run_all_in_graph_and_eager_modes
class ZeroInflatedTest(tf.test.TestCase):

  def testP(self):
    p = [0.2, 0.4]
    dist = tfd.Bernoulli(probs=p)
    self.assertAllClose(p, self.evaluate(dist.probs))

  def testInvalidP(self):
    invalid_ps = [1.01, 2.]
    for p in invalid_ps:
      with self.assertRaisesOpError("probs has components greater than 1"):
        dist = tfd.Bernoulli(probs=p, validate_args=True)
        self.evaluate(dist.probs)

    invalid_ps = [-0.01, -3.]
    for p in invalid_ps:
      with self.assertRaisesOpError("Condition x >= 0"):
        dist = tfd.Bernoulli(probs=p, validate_args=True)
        self.evaluate(dist.probs)

    valid_ps = [0.0, 0.5, 1.0]
    for p in valid_ps:
      dist = tfd.Bernoulli(probs=p)
      self.assertEqual(p, self.evaluate(dist.probs))  # Should not fail

  def testShapes(self):
    for batch_shape in ([], [1], [2, 3, 4]):
      dist = make_bernoulli(batch_shape)
      self.assertAllEqual(batch_shape, dist.batch_shape.as_list())
      self.assertAllEqual(batch_shape, self.evaluate(dist.batch_shape_tensor()))
      self.assertAllEqual([], dist.event_shape.as_list())
      self.assertAllEqual([], self.evaluate(dist.event_shape_tensor()))

  def testDtype(self):
    dist = make_bernoulli([])
    self.assertEqual(dist.dtype, tf.int32)
    self.assertEqual(dist.dtype, dist.sample(5).dtype)
    self.assertEqual(dist.dtype, dist.mode().dtype)
    self.assertEqual(dist.probs.dtype, dist.mean().dtype)
    self.assertEqual(dist.probs.dtype, dist.variance().dtype)
    self.assertEqual(dist.probs.dtype, dist.stddev().dtype)
    self.assertEqual(dist.probs.dtype, dist.entropy().dtype)
    self.assertEqual(dist.probs.dtype, dist.prob(0).dtype)
    self.assertEqual(dist.probs.dtype, dist.prob(0.5).dtype)
    self.assertEqual(dist.probs.dtype, dist.log_prob(0).dtype)
    self.assertEqual(dist.probs.dtype, dist.log_prob(0.5).dtype)

    dist64 = make_bernoulli([], tf.int64)
    self.assertEqual(dist64.dtype, tf.int64)
    self.assertEqual(dist64.dtype, dist64.sample(5).dtype)
    self.assertEqual(dist64.dtype, dist64.mode().dtype)

  def _testPmf(self, **kwargs):
    dist = tfd.Bernoulli(**kwargs)
    # pylint: disable=bad-continuation
    xs = [
        0,
        [1],
        [1, 0],
        [[1, 0]],
        [[1, 0], [1, 1]],
    ]
    expected_pmfs = [
        [[0.8, 0.6], [0.7, 0.4]],
        [[0.2, 0.4], [0.3, 0.6]],
        [[0.2, 0.6], [0.3, 0.4]],
        [[0.2, 0.6], [0.3, 0.4]],
        [[0.2, 0.6], [0.3, 0.6]],
    ]
    # pylint: enable=bad-continuation

    for x, expected_pmf in zip(xs, expected_pmfs):
      self.assertAllClose(self.evaluate(dist.prob(x)), expected_pmf)
      self.assertAllClose(self.evaluate(dist.log_prob(x)), np.log(expected_pmf))

  def testPmfCorrectBroadcastDynamicShape(self):
    p = tf.compat.v1.placeholder_with_default([0.2, 0.3, 0.4], shape=None)
    dist = tfd.Bernoulli(probs=p)
    event1 = [1, 0, 1]
    event2 = [[1, 0, 1]]
    self.assertAllClose(
        [0.2, 0.7, 0.4], self.evaluate(dist.prob(event1)))
    self.assertAllClose(
        [[0.2, 0.7, 0.4]], self.evaluate(dist.prob(event2)))

  def testPmfInvalid(self):
    p = [0.1, 0.2, 0.7]
    dist = tfd.Bernoulli(probs=p, validate_args=True)
    with self.assertRaisesOpError("must be non-negative."):
      self.evaluate(dist.prob([1, 1, -1]))
    with self.assertRaisesOpError("Elements cannot exceed 1."):
      self.evaluate(dist.prob([2, 0, 1]))

  def testPmfWithP(self):
    p = [[0.2, 0.4], [0.3, 0.6]]
    self._testPmf(probs=p)
    if not special:
      return
    self._testPmf(logits=special.logit(p))

  def testPmfWithFloatArgReturnsXEntropy(self):
    p = [[0.2], [0.4], [0.3], [0.6]]
    samps = [0, 0.1, 0.8]
    self.assertAllClose(
        np.float32(samps) * np.log(np.float32(p)) +
        (1 - np.float32(samps)) * np.log(1 - np.float32(p)),
        self.evaluate(
            tfd.Bernoulli(probs=p, validate_args=False).log_prob(samps)))

  def testBroadcasting(self):
    probs = lambda p: tf.compat.v1.placeholder_with_default(p, shape=None)
    dist = lambda p: tfd.Bernoulli(probs=probs(p))
    self.assertAllClose(np.log(0.5), self.evaluate(dist(0.5).log_prob(1)))
    self.assertAllClose(
        np.log([0.5, 0.5, 0.5]), self.evaluate(dist(0.5).log_prob([1, 1, 1])))
    self.assertAllClose(np.log([0.5, 0.5, 0.5]),
                        self.evaluate(dist([0.5, 0.5, 0.5]).log_prob(1)))

  def testPmfShapes(self):
    probs = lambda p: tf.compat.v1.placeholder_with_default(p, shape=None)
    dist = lambda p: tfd.Bernoulli(probs=probs(p))
    self.assertEqual(
        2, len(self.evaluate(dist([[0.5], [0.5]]).log_prob(1)).shape))

    dist = tfd.Bernoulli(probs=0.5)
    self.assertEqual(2, len(self.evaluate(dist.log_prob([[1], [1]])).shape))

    dist = tfd.Bernoulli(probs=0.5)
    self.assertEqual((), dist.log_prob(1).shape)
    self.assertEqual((1), dist.log_prob([1]).shape)
    self.assertEqual((2, 1), dist.log_prob([[1], [1]]).shape)

    dist = tfd.Bernoulli(probs=[[0.5], [0.5]])
    self.assertEqual((2, 1), dist.log_prob(1).shape)

  def testBoundaryConditions(self):
    dist = tfd.Bernoulli(probs=1.0)
    self.assertAllClose(np.nan, self.evaluate(dist.log_prob(0)))
    self.assertAllClose([np.nan], [self.evaluate(dist.log_prob(1))])

  def testEntropyNoBatch(self):
    p = 0.2
    dist = tfd.Bernoulli(probs=p)
    self.assertAllClose(self.evaluate(dist.entropy()), entropy(p))

  def testEntropyWithBatch(self):
    p = [[0.1, 0.7], [0.2, 0.6]]
    dist = tfd.Bernoulli(probs=p, validate_args=False)
    self.assertAllClose(
        self.evaluate(dist.entropy()),
        [[entropy(0.1), entropy(0.7)], [entropy(0.2),
                                        entropy(0.6)]])

  def testSampleN(self):
    p = [0.2, 0.6]
    dist = tfd.Bernoulli(probs=p)
    n = 100000
    samples = dist.sample(n)
    samples.set_shape([n, 2])
    self.assertEqual(samples.dtype, tf.int32)
    sample_values = self.evaluate(samples)
    self.assertTrue(np.all(sample_values >= 0))
    self.assertTrue(np.all(sample_values <= 1))
    # Note that the standard error for the sample mean is ~ sqrt(p * (1 - p) /
    # n). This means that the tolerance is very sensitive to the value of p
    # as well as n.
    self.assertAllClose(p, np.mean(sample_values, axis=0), atol=1e-2)
    self.assertEqual(set([0, 1]), set(sample_values.flatten()))
    # In this test we're just interested in verifying there isn't a crash
    # owing to mismatched types. b/30940152
    dist = tfd.Bernoulli(np.log([.2, .4]))
    self.assertAllEqual((1, 2), dist.sample(
        1, seed=tfp_test_util.test_seed()).shape.as_list())

  def testNotReparameterized(self):
    p = tf.constant([0.2, 0.6])
    _, grad_p = tfp.math.value_and_gradient(
        lambda x: tfd.Bernoulli(probs=x).sample(100), p)
    self.assertIsNone(grad_p)

  def testSampleDeterministicScalarVsVector(self):
    p = [0.2, 0.6]
    dist = tfd.Bernoulli(probs=p)
    n = 1000

    def _maybe_seed():
      if tf.executing_eagerly():
        tf.compat.v1.set_random_seed(42)
        return None
      return 42
    self.assertAllEqual(
        self.evaluate(dist.sample(n, _maybe_seed())),
        self.evaluate(dist.sample([n], _maybe_seed())))
    n = tf.compat.v1.placeholder_with_default(np.int32(1000), shape=None)
    sample1 = dist.sample(n, _maybe_seed())
    sample2 = dist.sample([n], _maybe_seed())
    sample1, sample2 = self.evaluate([sample1, sample2])
    self.assertAllEqual(sample1, sample2)

  def testMean(self):
    p = np.array([[0.2, 0.7], [0.5, 0.4]], dtype=np.float32)
    dist = tfd.Bernoulli(probs=p)
    self.assertAllEqual(self.evaluate(dist.mean()), p)

  def testVarianceAndStd(self):
    var = lambda p: p * (1. - p)
    p = [[0.2, 0.7], [0.5, 0.4]]
    dist = tfd.Bernoulli(probs=p)
    self.assertAllClose(
        self.evaluate(dist.variance()),
        np.array([[var(0.2), var(0.7)], [var(0.5), var(0.4)]],
                 dtype=np.float32))
    self.assertAllClose(
        self.evaluate(dist.stddev()),
        np.array([[np.sqrt(var(0.2)), np.sqrt(var(0.7))],
                  [np.sqrt(var(0.5)), np.sqrt(var(0.4))]],
                 dtype=np.float32))

  def testBernoulliBernoulliKL(self):
    batch_size = 6
    a_p = np.array([0.5] * batch_size, dtype=np.float32)
    b_p = np.array([0.4] * batch_size, dtype=np.float32)

    a = tfd.Bernoulli(probs=a_p)
    b = tfd.Bernoulli(probs=b_p)

    kl = tfd.kl_divergence(a, b)
    kl_val = self.evaluate(kl)

    kl_expected = (a_p * np.log(a_p / b_p) + (1. - a_p) * np.log(
        (1. - a_p) / (1. - b_p)))

    self.assertEqual(kl.shape, (batch_size,))
    self.assertAllClose(kl_val, kl_expected)

if __name__ == "__main__":
  tf.test.main()



# This script generate synthetic data
from __future__ import print_function, division, absolute_import

import numpy as np

from odin.fuel.dataset import Dataset

# ===========================================================================
# Helpers
# ===========================================================================
def rand_probabilities(p_min=0. + 10e-8, p_max=1. - 10e-8,
                       shape=(), seed=None):
  if isinstance(seed, np.random.RandomState):
    rand = seed
  else:
    rand = np.random.RandomState(seed=seed)
  return rand.rand(*shape) * (p_max - p_min) + p_min

def rand_samples_per_distribution(n_total, n_dist, seed=None):
  if isinstance(seed, np.random.RandomState):
    rand = seed
  else:
    rand = np.random.RandomState(seed=seed)
  if n_dist == 1:
    n_samples_per_distribution = (n_total,)
  else: # stick breaking
    n_samples_per_distribution = sorted(rand.choice(
        a=np.arange(n_total), size=n_dist - 1, replace=False)) + [n_total]
    n_samples_per_distribution = (n_samples_per_distribution[0],) + \
    tuple([i1 - i0
        for i0, i1, in zip(n_samples_per_distribution, n_samples_per_distribution[1:])])
  assert np.sum(n_samples_per_distribution) == n_total
  return n_samples_per_distribution

# ===========================================================================
# Synthesizing datasets
# ===========================================================================
def generate_ZeroInflatedNegativeBinomial(
    n_samples=12000, n_features=8, n_labels=3,
    n=[5, 25], p_success=[0.1, 0.5], p_drop=[0.2, 0.5],
    seed=52181208):
  rand = np.random.RandomState(seed)
  n_samples_per_distribution = rand_samples_per_distribution(
      n_total=n_samples + 1000, n_dist=n_labels)
  # ====== sampling ====== #
  X = []
  y = []
  params = []
  for i in range(n_labels):
    # random parameters
    shape = (n_samples_per_distribution[i], n_features)
    total_count = rand.randint(n[0], n[1] + 1)
    prob_sucesss = rand_probabilities(p_min=p_success[0],
                                      p_max=p_success[1],
                                      seed=rand)
    prob_keep = rand_probabilities(p_min=1. - p_drop[1],
                                   p_max=1. - p_drop[0],
                                   seed=rand)
    params.append((total_count, prob_sucesss, 1 - prob_keep))
    # random data
    data = rand.negative_binomial(n=total_count, p=prob_sucesss, size=shape)
    mask = np.random.binomial(n=1, p=prob_keep, size=shape)
    newdata = (data * mask)
    indices_to_keep = (newdata.sum(axis=1) > 0).ravel()
    newdata = newdata[indices_to_keep]
    X.append(newdata)
    y += [i] * shape[0]
  # ====== post processing ====== #
  X = np.concatenate(X, axis=0)
  y = np.array(y)
  ids = rand.permutation(X.shape[0])[:n_samples]
  X = X[ids]
  y = y[ids]
  assert X.shape[0] == y.shape[0] == n_samples
  return X, y, params

def generate_ZeroInflatedPoisson(
    n_samples=12000, n_features=8, n_labels=3,
    mu=[5, 25], theta=[2, 8], p_drop=[0.2, 0.5],
    seed=52181208):
  rand = np.random.RandomState(seed)
  n_samples_per_distribution = rand_samples_per_distribution(
      n_total=n_samples + 1000, n_dist=n_labels)
  # ====== sampling ====== #
  # mean Gamma: shape * scale
  X = []
  y = []
  params = []
  for i in range(n_labels):
    # random parameters
    shape = (n_samples_per_distribution[i], n_features)
    mu_ = rand.randint(mu[0], mu[1] + 1)
    theta_ = rand.randint(theta[0], theta[1] + 1)
    p = mu_ / (mu_ + theta_)
    r = theta_
    prob_keep = rand_probabilities(p_min=1. - p_drop[1],
                                   p_max=1. - p_drop[0],
                                   seed=rand)
    params.append((mu_, theta_, 1 - prob_keep))
    # random data
    data = np.random.poisson(
        np.random.gamma(r, p / (1 - p), size=shape)
    )
    mask = np.random.binomial(n=1, p=prob_keep, size=shape)
    newdata = (data * mask)
    indices_to_keep = (newdata.sum(axis=1) > 0).ravel()
    newdata = newdata[indices_to_keep]
    X.append(newdata)
    y += [i] * shape[0]
  # ====== post processing ====== #
  X = np.concatenate(X, axis=0)
  y = np.array(y)
  ids = rand.permutation(X.shape[0])[:n_samples]
  X = X[ids]
  y = y[ids]
  assert X.shape[0] == y.shape[0] == n_samples
  return X, y, params



from __future__ import division, absolute_import, print_function
from numbers import Number

import numpy as np
import tensorflow as tf

from odin import backend as K
from odin.nnet.base import NNOp
from odin.utils import is_number, is_string, as_tuple

def _preprocess_windows(window, ndims):
  if len(window) == ndims - 2:
    return window
  else:
    return as_tuple(window, N=(ndims - 2))

class Pool(NNOp):
  """
  Parameters
  ----------
  pool_size : tuple of int or just an int.
      Factor by which to downscale (vertical ws, horizontal ws, ...).
  strides : tuple of two ints or theano vector of ints of size 2.
      Stride size, which is the number of shifts over rows/cols to get the
      next pool region.
      If stride is None, it is considered equal to pool_size (no overlap
      on pooling regions).
  pad : tuple of two ints or theano vector of ints of size 2.
      (pad_h, pad_w), pad zeros to extend beyond four borders of the
      images, pad_h is the size of the top and bottom margins, and
      pad_w is the size of the left and right margins.
  mode : {'max', 'avg'}
      Operation executed on each window. `max` or `average`
  pool_func : 'auto' or call-able
      if 'auto', auto select pool function based on number of input
      dimension (pool2D for 4D input, pool3D for 5D input)
  transpose_mode: 'nn', 'pad', 'pad_margin', 'repeat'
      One of the mode for `transposing` downsampling process,
      check `odin.nnet.sampling.Upsampling`.

  Note
  ----
  This pooling algorithm has non-deterministic behaviour on cuDNN
  """

  def __init__(self, pool_size=2, strides=None, dilation=1,
               pad='valid', mode='max', transpose_mode='nn', **kwargs):
    super(Pool, self).__init__(**kwargs)
    self.pool_size = as_tuple(pool_size, t=int)
    self.strides = self.pool_size if strides is None \
        else as_tuple(strides, t=int)
    self.dilation = (1,) if dilation is None else as_tuple(dilation, t=int)
    self.pad = pad.upper() if is_string(pad) else as_tuple(pad, t=int)
    self.mode = mode.upper()
    self.transpose_mode = transpose_mode

  def _apply(self, X):
    ndims = X.shape.ndims
    return tf.nn.pool(X,
        window_shape=_preprocess_windows(self.pool_size, ndims),
        strides=_preprocess_windows(self.strides, ndims),
        dilation_rate=_preprocess_windows(self.dilation, ndims),
        padding=self.pad, pooling_type=self.mode)

  def _transpose(self):
    return Upsample(size=self.strides, axes='auto',
        mode=self.transpose_mode, transpose_mode=self.mode,
        desire_shape=self.input_shape)

class Upsample(NNOp):
  """ Upsampling

  Parameters
  ----------
  size: int
      upsampling size (new_size = input_size * size)
  axes: int, list of int
      the axes of tensor which the upsampling method will be applied
  mode: str, int
      'nn' for nearest neighbor (e.g. [1, 2] => [1, 1, 2, 2]),
      'pad' for padding within the tensor. 'pad_margin' do padding
      in the margin of the tensor. 'repeat' simple algorithm for
      repeating the element (e.g. [1, 2] => [1, 2, 1, 2])
  transpose_mode: {'max', 'avg'}
      pass
  desire_shape : shape tuple
      specific output shape, if the upsampled image is bigger, then crop
      the image, any axis specified with negative value or `None` will
      be kept unchanged
  """

  def __init__(self, size=2, axes='auto', mode='nn', transpose_mode='max',
               desire_shape=None, **kwargs):
    super(Upsample, self).__init__(**kwargs)
    self.size = size
    self.axes = axes
    self.mode = mode
    self.transpose_mode = transpose_mode
    self.desire_shape = desire_shape

  def _apply(self, X):
    axes = self.axes
    ndims = X.shape.ndims
    if is_string(axes) and axes.lower() == 'auto':
      if ndims == 3:
        axes = (1,)
      elif ndims == 4:
        axes = (1, 2)
      elif ndims == 5:
        axes = (1, 2, 3)
    X = K.upsample(X, scale=self.size, axes=axes, method=self.mode)
    # ====== check desire_shape ====== #
    desire_shape = self.desire_shape
    if desire_shape is not None:
      desire_shape = [None if i is None or i < 0 else int(i)
                      for i in desire_shape]
      # do padding if necessary
      paddings = [[0, 0] if i is None or o is None or i >= o else
                  [tf.cast(tf.ceil((o - i) / 2), 'int32'),
                   tf.cast(tf.floor((o - i) / 2), 'int32')]
                  for i, o in zip(X.shape.as_list(), desire_shape)]
      if not all(i == [0, 0] for i in paddings):
        X = tf.pad(X, paddings=paddings, mode='CONSTANT')
      # do slice if necessary
      slices = [slice(tf.cast(tf.floor((i - o) / 2), 'int32'),
                      tf.cast(-tf.ceil((i - o) / 2), 'int32'), None)
                if i is not None and o is not None and i > o else slice(None)
                for i, o in zip(X.shape.as_list(), desire_shape)]
      if any(s is not slice(None) for s in slices):
        X = X[slices]
      K.set_shape(X, tuple([i if is_number(i) else None
                            for i in desire_shape]))
    return X

  def _transpose(self):
    if self.axes != 'auto':
      raise RuntimeError("Do not support tranpose of Upsample with "
                         "axes=%s, the only support value is 'auto'."
                         % self.axes)
    return Pool(pool_size=self.size, strides=None, pad='valid',
        mode=self.transpose_mode, transpose_mode=self.mode)

class StatsPool(NNOp):
  """ Calculate mean and stddev of input Tensor along `axes`
  then merge the statistics based on given `output_mode`

  This class is for parsing lines like
    stats-layer name=tdnn1-stats config=mean+stddev(-99:3:9:99) input=tdnn1
    This adds statistics-pooling and statistics-extraction components.  An
    example string is 'mean(-99:3:9::99)', which means, compute the mean of
    data within a window of -99 to +99, with distinct means computed every 9
    frames (we round to get the appropriate one), and with the input extracted
    on multiples of 3 frames (so this will force the input to this layer to be
    evaluated every 3 frames).  Another example string is
    'mean+stddev(-99:3:9:99)', which will also cause the standard deviation to
    be computed.
    The dimension is worked out from the input. mean and stddev add a
    dimension of input_dim each to the output dimension. If counts is
    specified, an additional dimension is added to the output to store log
    counts.
  """
  _SUPPORT_OUTPUT_MODE = ('concat', 'sum', 'mean', 'max')

  def __init__(self, axes, output_mode='concat', **kwargs):
    super(StatsPool, self).__init__(**kwargs)
    self.axes = axes
    # ====== check output mode ====== #
    output_mode = str(output_mode).lower()
    if output_mode not in StatsPool._SUPPORT_OUTPUT_MODE:
      raise ValueError("No support for `output_mode=%s`, the following values "
                       "are accepted: %s" %
                       (output_mode, StatsPool._SUPPORT_OUTPUT_MODE))
    self.output_mode = output_mode

  def _apply(self, X):
    mean, var = tf.nn.moments(x=X, axes=self.axes, keep_dims=True)
    std = tf.sqrt(var)
    if self.output_mode == 'concat':
      X = tf.concat((mean, std), self.axes)
    elif self.output_mode == 'sum':
      X = mean + std
    elif self.output_mode == 'mean':
      X = (mean + std) / 2
    elif self.output_mode == 'max':
      X = tf.maximum(mean, std)
    return X



# -*- coding: utf-8 -*-
# This module contain the implementation of Time-delayed neural network
# based on variate source, using both Dense (classic) and Convolution (modern)
# for modeling delayed inputs.
#
# The inputs could be 3-D or 4-D tensor
#  [n_samples, n_timestep, n_features]
# or
#  [n_samples, n_timestep, n_features, n_channels]
#
# References
# ----------
# [1] Waibel, A., Hanazawa, T., Hinton, G., Shikano, K., & Lang, K. J. (1989).
# Phoneme recognition using time-delay neural networks. IEEE transactions on
# acoustics, speech, and signal processing, 37(3), 328-339.
# [2] Peddinti, V., Povey, D., & Khudanpur, S. (2015). A time delay neural
# network architecture for efficient modeling of long temporal contexts.
# In INTERSPEECH (pp. 3214-3218).
# [3] Yoon Kim, Yacine Jernite, David Sontag, and Alexander M. Rush. 2016.
# Character-aware neural language models, AAAI'16
from __future__ import print_function, division, absolute_import

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin import backend as K
from odin.autoconfig import randint
from odin.nnet.base import NNOp
from odin.utils import as_tuple, is_string
from odin.backend.role import Weight, Bias, ConvKernel

_allow_time_pool = ('min', 'max', 'avg', 'sum', 'stat', 'none')

# ===========================================================================
# Time-Delayed Dense
# ===========================================================================
class TimeDelayedDense(NNOp):
  """ TimeDelayedDense

  Parameters
  ----------
  n_new_features : {None, integer, list of integer}
      the amount of weighted sum (Dense layer) will be performed for each
      time slice of `n_time_context`

      if None, or empty list, no projection is performed.
      Otherwise, a shallow Dense network or deep Dense network can be
      specified by a list of integer representing number of hidden unit.

  n_time_context : int
      the length of time window context, i.e. kernel size in
      time dimension for convolution operator

  time_pool : {None, 'max', 'avg', 'stat'}
      pooling in time dimension for each time slice (i.e. different from
      `TimeDelayedConv`, when the pooling is performed after computed
      all time sliced)

      for 'stat' pooling, mean and standard deviation is calculated along
      time-dimension, then output the concatenation of the two.

      if None, concatenate all the output of time slice into a vector,
      which return a 3-D tensor with shape
      [n_sample, n_timestep - n_time_context + 1, n_new_features[-1] * n_time_context]

  """

  def __init__(self, n_new_features, n_time_context,
               time_pool='max', backward=False,
               W_init=init_ops.glorot_uniform_initializer(seed=randint()),
               b_init=init_ops.constant_initializer(0),
               activation=K.linear, **kwargs):
    super(TimeDelayedDense, self).__init__(**kwargs)
    if n_new_features is None:
      self.n_new_features = []
    else:
      self.n_new_features = as_tuple(n_new_features, t=int)
    self.n_time_context = int(n_time_context)
    self.n_layers = len(self.n_new_features)
    # ====== initialization ====== #
    self.W_init = W_init
    self.b_init = b_init
    # ====== activation ====== #
    if activation is None:
      activation = K.linear
    if not isinstance(activation, (tuple, list)):
      activation = (activation,)
    activation = [K.linear if i is None else i
                  for i in activation]
    self.activation = as_tuple(activation, N=self.n_layers)
    # ====== time axis manipulation ====== #
    time_pool = str(time_pool).lower()
    assert time_pool in _allow_time_pool, \
    "Only support: %s; but given: '%s'" % (str(_allow_time_pool), str(time_pool))
    self.time_pool = time_pool
    self.backward = bool(backward)

  def _initialize(self):
    time_dim = self.input_shape[1]
    feat_dim = self.input_shape[2]
    layers = [feat_dim] + list(self.n_new_features)
    for i, (l1, l2) in enumerate(zip(layers, layers[1:])):
      # weights
      self.get_variable_nnop(initializer=self.W_init,
          shape=(l1, l2),
          name='W%d' % i, roles=Weight)
      # biases
      if self.b_init is not None:
        self.get_variable_nnop(initializer=self.b_init,
            shape=(l2,), name='b%d' % i, roles=Bias)

  def _apply(self, X):
    if X.shape.ndims >= 4:
      X = tf.reshape(X, shape=(X.shape[0], X.shape[1], -1))
    assert X.shape.ndims == 3, \
    "TimeDelayedConv require 3-D or 4-D input, but given: %s" % str(X)
    # [n_sample, n_timestep, n_features]
    time_dim = X.shape[1].value
    feat_dim = X.shape[2].value
    new_time_dim = time_dim - self.n_time_context + 1
    new_feat_dim = (self.n_new_features[-1]
                    if len(self.n_new_features) > 0 else
                    feat_dim)
    # ====== traverse backward along time axis ====== #
    if self.backward:
      X = tf.reverse(X, 1)
    # ====== prepare VALID padding ====== #
    context_indices = tf.range(0, new_time_dim, delta=1,
                               dtype=tf.int32, name="ContextIndices")
    if self.time_pool in ('concat', 'none'):
      init_dim = new_feat_dim * self.n_time_context
    elif self.time_pool in ('max', 'avg', 'min', 'sum'):
      init_dim = new_feat_dim
    elif self.time_pool == 'stat':
      init_dim = new_feat_dim * 2
    initializer = tf.zeros(
        shape=(tf.shape(X)[0], init_dim),
        dtype=self.get('W0').dtype if len(self.n_new_features) > 0 else X.dtype)

    # ====== step function ====== #
    def _time_step(o, ids):
      ctx = X[:, ids:ids + self.n_time_context, :]
      ctx.set_shape((ctx.shape[0], self.n_time_context, ctx.shape[2]))
      ctx = tf.reshape(ctx, shape=(-1, ctx.shape[2]))
      # applying deep dense network
      for l in range(self.n_layers):
        ctx = K.dot(ctx, self.get('W%d' % l))
        if self.b_init is not None:
          ctx = ctx + self.get('b%d' % l)
        ctx = self.activation[l](ctx)
      ctx = tf.reshape(ctx, shape=(-1, self.n_time_context, new_feat_dim))
      # applying pooling
      if self.time_pool in ('concat', 'none'):
        ctx = tf.reshape(ctx,
                         shape=(tf.shape(ctx)[0], self.n_time_context * new_feat_dim))
      elif self.time_pool == 'max':
        ctx = tf.reduce_max(ctx, axis=1)
      elif self.time_pool == 'min':
        ctx = tf.reduce_min(ctx, axis=1)
      elif self.time_pool == 'sum':
        ctx = tf.reduce_sum(ctx, axis=1)
      elif self.time_pool == 'avg':
        ctx = tf.reduce_mean(ctx, axis=1)
      elif self.time_pool == 'stat':
        mean, var = tf.nn.moments(ctx, axes=1)
        ctx = tf.concat([mean, tf.sqrt(var)], -1)
      return ctx
    # ====== apply convolution ====== #
    output = K.scan_tensors(fn=_time_step,
                            sequences=context_indices,
                            mask=None,
                            initializer=initializer, axis=0,
                            parallel_iterations=12)
    output = tf.transpose(output, perm=(1, 0, 2))
    # [n_sample, n_timestep - n_time_context + 1, n_new_features]
    return output

# ===========================================================================
# Time-Delayed Convolution
# ===========================================================================
class TimeDelayedConv(NNOp):
  """ Time-delayed convolutional neural network
  Input is 3-D tensor `[n_sample, n_timestep, n_features]`
  Output is 2-D time-pooled tensor `[n_sample, n_new_features]`

  From the paper, it is suggested to create multiple `TimeDelayedConv`
  with variate number of feature map and length of context windows,
  then concatenate the outputs for `Dense` layers

  For example:
   - feature_maps = [50, 100, 150, 200, 200, 200, 200]
   - kernels = [1, 2, 3, 4, 5, 6, 7]

  Parameters
  ----------
  n_new_features : int
      The number of learn-able convolutional filters this layer has.

  n_time_context : int
      the length of time window context, i.e. kernel size in
      time dimension for convolution operator

  time_pool : {None, 'max', 'avg', 'stat'}
      pooling in time dimension after convolution operator
      for 'stat' pooling, mean and standard deviation is calculated along
      time-dimension, then output the concatenation of the two.
      if None, no pooling is performed, the output is returned in
      shape [n_samples, n_reduced_timestep, n_new_features]

  """

  def __init__(self, n_new_features, n_time_context,
               time_pool='max', backward=False,
               W_init=init_ops.glorot_uniform_initializer(seed=randint()),
               b_init=init_ops.constant_initializer(0),
               activation=K.linear, **kwargs):
    super(TimeDelayedConv, self).__init__(**kwargs)
    self.n_new_features = int(n_new_features)
    self.n_time_context = int(n_time_context)
    self.W_init = W_init
    self.b_init = b_init
    self.activation = activation
    # ====== time axis manipulation ====== #
    time_pool = str(time_pool).lower()
    assert time_pool in _allow_time_pool, \
    "Only support: %s; but given: '%s'" % (str(_allow_time_pool), str(time_pool))
    self.time_pool = time_pool
    self.backward = bool(backward)

  def _initialize(self):
    time_dim = self.input_shape[1]
    feat_dim = self.input_shape[2]
    channel_dim = 1 if len(self.input_shape) == 3 else self.input_shape[-1]
    # weights
    self.get_variable_nnop(initializer=self.W_init,
        shape=(self.n_time_context, feat_dim, channel_dim, self.n_new_features),
        name='W', roles=ConvKernel)
    if self.b_init is not None:
      self.get_variable_nnop(initializer=self.b_init,
          shape=(self.n_new_features,), name='b', roles=Bias)

  def _apply(self, X):
    if X.shape.ndims == 3:
      X = tf.expand_dims(X, axis=-1)
    assert X.shape.ndims == 4, \
    "TimeDelayedConv require 3-D or 4-D input, but given: '%s'" % str(X)
    # [n_sample, n_timestep, n_features, 1]
    # ====== traverse backward along time axis ====== #
    if self.backward:
      X = tf.reverse(X, 1)
    # ====== apply convolution ====== #
    conved = tf.nn.convolution(input=X, filter=self.get('W'),
        padding="VALID",
        strides=[1, 1],
        data_format="NHWC")
    # [n_sample, n_timestep - n_time_context + 1, 1, n_new_features]
    # ====== apply bias ====== #
    if 'b' in self.variable_info:
      conved += K.dimshuffle(self.get('b'), ('x', 'x', 'x', 0))
    # ====== activation ====== #
    conved = self.activation(conved)
    # ====== applying pooling ====== #
    if self.time_pool == 'none':
      pool = tf.squeeze(conved, 2)
      # [n_sample, n_timestep - n_time_context + 1, n_new_features]
    elif self.time_pool == 'stat':
      mean, var = tf.nn.moments(conved, axes=1, keep_dims=True)
      std = tf.sqrt(var)
      pool = tf.concat([mean, std], -1)
      pool = tf.squeeze(pool, axis=[1, 2])
      # [n_sample, n_new_features * 2]
    elif self.time_pool in ('avg', 'max'):
      fn_pool = tf.nn.max_pool if self.time_pool == 'max' else tf.nn.avg_pool
      pool = fn_pool(conved,
                     ksize=[1, tf.shape(conved)[1], 1, 1],
                     strides=[1, 1, 1, 1],
                     padding='VALID',
                     data_format="NHWC")
      pool = tf.squeeze(pool, axis=[1, 2])
      # [n_sample, n_new_features]
    elif self.time_pool in ('sum'):
      pool = tf.reduce_mean(conved, axis=1)
      pool = tf.squeeze(pool, axis=1)
    # ====== return 2D output ====== #
    return pool



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

import os
os.environ['ODIN'] = 'gpu,float32'
import shutil
from collections import defaultdict

import numpy as np
import numba as nb
import tensorflow as tf
import seaborn as sns
import pandas as pd

from sklearn.metrics import accuracy_score, log_loss, f1_score

from odin import fuel as F
from odin import nnet as N, backend as K
from odin import visual as V
from odin.utils import (ctext, mpi, Progbar, catch_warnings_ignore, stdio,
                        get_logpath, catch_warnings_ignore)


from helpers import (FEATURE_RECIPE, FEATURE_NAME, PATH_ACOUSTIC_FEATURES,
                     MINIMUM_UTT_DURATION, ANALYSIS_DIR, Config,
                     filter_utterances, prepare_dnn_data)

# ====== prepare log ====== #
stdio(get_logpath(name="analyze_data.log", increasing=True,
                  odin_base=False, root=ANALYSIS_DIR))
print(ctext(FEATURE_RECIPE, 'lightyellow'))
print(ctext(FEATURE_NAME, 'lightyellow'))
assert os.path.isdir(os.path.join(PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE))
# ====== essential path ====== #
figure_path = os.path.join(ANALYSIS_DIR, '%s_%s.pdf' %
                           (FEATURE_RECIPE.replace('_', ''), FEATURE_NAME))
print(ctext(figure_path, 'lightyellow'))
# ===========================================================================
# Load the data
# ===========================================================================
ds = F.Dataset(os.path.join(PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE),
               read_only=True)
X = ds[FEATURE_NAME]
# remove all noise data
indices = {name: (start, end)
           for name, (start, end) in ds['indices_%s' % FEATURE_NAME].items()
           if '/' not in name}

all_dataset = sorted(set(ds['dsname'].values()))
print("All dataset:", ctext(all_dataset, 'cyan'))
# ===========================================================================
# Helpers
# ===========================================================================
all_percentiles = [.01, .05, .1, .25, .5, .75, .9, .95, .99]

def clipping_quartile(x, alpha=8):
  # For removing outliers
  # result wider range threshold, keep more data points,
  x = x.astype('int32')
  Q1 = np.percentile(x, q=25)
  Q3 = np.percentile(x, q=75)
  IQR = Q3 - Q1
  x = x[x < Q3 + alpha * IQR]
  x = x[x > Q1 - alpha * IQR]
  return x

def dataset_statistics(dsname):
  ids = {name: (start, end)
         for name, (start, end) in indices.items()
         if ds['dsname'][name] == dsname}
  name2spk = {name: ds['spkid'][name]
              for name in ids.keys()}
  s = []
  s.append('=' * 12 + ctext('%-12s' % dsname, 'lightyellow') + '=' * 12)
  s.append('#Files   :' + ctext(len(ids), 'cyan'))
  s.append("#Speakers:" + ctext(len(set(name2spk.values())), 'cyan'))
  # ====== mean and std ====== #
  sum1 = 0.
  sum2 = 0.
  n = 0
  spk_sum1 = defaultdict(float)
  spk_sum2 = defaultdict(float)
  spk_n = defaultdict(int)
  for name, (start, end) in ids.items():
    spkid = name2spk[name]
    n += end - start; spk_n[spkid] += end - start
    x = X[start:end][:].astype('float64')
    s1 = np.sum(x, axis=0); s2 = np.sum(x**2, axis=0)
    sum1 += s1; sum2 += s2
    spk_sum1[spkid] += s1; spk_sum2[spkid] += s2
  data_mean = sum1 / n
  data_std = np.sqrt(sum2 / n - data_mean ** 2)

  spk_stats = {}
  for spkid in name2spk.values():
    n = spk_n[spkid]
    s1, s2 = spk_sum1[spkid], spk_sum2[spkid]
    mean = s1 / n
    std = np.sqrt(s2 / n - mean ** 2)
    spk_stats[spkid] = (mean, std)
  spk_mean = np.concatenate([x[0][None, :] for x in spk_stats.values()],
                            axis=0).mean(0)
  spk_std = np.concatenate([x[1][None, :] for x in spk_stats.values()],
                           axis=0).mean(0)
  # ====== utterances length ====== #
  all_length = np.array([(end - start) * Config.STEP_LENGTH
                         for start, end in indices.values()])
  # ====== speaker - utterance relation ====== #
  nutt_per_spk = defaultdict(int)
  dur_per_spk = defaultdict(list)
  for name, (start, end) in ids.items():
    spkid = name2spk[name]
    nutt_per_spk[spkid] += 1
    dur_per_spk[spkid].append((end - start) * Config.STEP_LENGTH)
  all_spk = sorted(nutt_per_spk.keys())
  spk_df = pd.DataFrame(data={'nutt_per_spk': [nutt_per_spk[spk] for spk in all_spk],
                              'sum_per_spk': [np.sum(dur_per_spk[spk]) for spk in all_spk],
                              'mean_per_spk': [np.mean(dur_per_spk[spk]) for spk in all_spk],
                        })
  return dsname, '\n'.join(s), (all_length, spk_df), (data_mean, data_std), (spk_mean, spk_std)

# ===========================================================================
# Create the report
# ===========================================================================
n_bin = 120
n_col = 3
linestyles = ['-', '--', '-.']

def plot_histogram(series, ax, title):
  V.plot_histogram(x=clipping_quartile(series), bins=n_bin, ax=ax,
                   title=title, fontsize=4)

def plot_mean_std(_map, title):
  V.plot_figure(nrow=6, ncol=20)
  for i, dsname in enumerate(all_dataset):
    mean, _ = _map[dsname]
    plt.plot(mean,
             linewidth=1.,
             linestyle=linestyles[i % len(linestyles)],
             label=dsname)
  plt.legend()
  plt.suptitle("[%s]Mean" % title)

  V.plot_figure(nrow=6, ncol=20)
  for i, dsname in enumerate(all_dataset):
    _, std = _map[dsname]
    plt.plot(std,
             linewidth=1.,
             linestyle=linestyles[i % len(linestyles)],
             label=dsname)
  plt.legend()
  plt.suptitle("[%s]StandardDeviation" % title)

with catch_warnings_ignore(RuntimeWarning), catch_warnings_ignore(FutureWarning):
  data_map = {}
  stats_map = {}
  spk_map = {}
  for dsname, text, data, stats, spk_stats in mpi.MPI(jobs=all_dataset, func=dataset_statistics,
                            ncpu=None, batch=1):
    data_map[dsname] = data
    stats_map[dsname] = stats
    spk_map[dsname] = spk_stats
    print(text)

  for dsname in all_dataset:
    print("Plotting ...", ctext(dsname, 'cyan'))
    data = data_map[dsname]
    V.plot_figure(nrow=2, ncol=20)
    ax = plt.subplot(1, n_col, 1)
    plot_histogram(data[0], ax, title="Duration")

    ax = plt.subplot(1, n_col, 2)
    plot_histogram(data[1]['sum_per_spk'], ax, title="Dur/Spk")

    ax = plt.subplot(1, n_col, 3)
    plot_histogram(data[1]['nutt_per_spk'], ax, title="#Utt/Spk")

    plt.suptitle(dsname, fontsize=8)

  plot_mean_std(_map=stats_map, title='Data')
  plot_mean_std(_map=spk_map, title='Speaker')

V.plot_save(figure_path, dpi=32)



from __future__ import division, absolute_import

from odin import backend as K
from odin.nnet.base import NNOp

import tensorflow as tf
from tensorflow.python.ops import init_ops


class Dropout(NNOp):
  """Computes dropout.

  With probability `keep_prob`, outputs the input element scaled up by
  `1 / keep_prob`, otherwise outputs `0`.  The scaling is so that the expected
  sum is unchanged.


  Parameters
  ----------
  x: A tensor.
      input tensor (any shape)
  level: float(0.-1.)
      probability dropout values in given tensor
  noise_dims: int or list(int)
      these dimensions will be setted to 1 in noise_shape, and
      used to broadcast the dropout mask.
  noise_type: 'gaussian' (or 'normal'), 'uniform'
      distribution used for generating noise
  rescale: bool
      whether rescale the outputs by dividing the retain probablity

  """

  def __init__(self, level=0.5,
               noise_dims=None, noise_type='uniform',
               rescale=True, **kwargs):
    super(Dropout, self).__init__(**kwargs)
    self.level = level
    self.noise_dims = noise_dims
    self.noise_type = noise_type
    self.rescale = rescale

  def _apply(self, X):
    return K.rand.apply_dropout(X, level=self.level,
                           noise_dims=self.noise_dims,
                           noise_type=self.noise_type,
                           rescale=self.rescale)

  def _transpose(self):
    return self


class Noise(NNOp):
  """
  Parameters
  ----------
  x: A tensor.
      input tensor (any shape)
  level : float or tensor scalar
      Standard deviation of added Gaussian noise, or range of
      uniform noise
  noise_dims: int or list(int)
      these dimensions will be setted to 1 in noise_shape, and
      used to broadcast the dropout mask.
  noise_type: 'gaussian' (or 'normal'), 'uniform'
      distribution used for generating noise

  """

  def __init__(self, level=0.075, noise_dims=None,
               noise_type='gaussian', **kwargs):
    super(Noise, self).__init__(**kwargs)
    self.level = level
    self.noise_dims = noise_dims
    self.noise_type = noise_type

  def _apply(self, X):
    # possible mechanism to hard override the noise-state
    return K.rand.apply_noise(X, level=self.level,
                         noise_dims=self.noise_dims,
                         noise_type=self.noise_type)

  def _transpose(self):
    return self


class GaussianDenoising(NNOp):
  """ Gaussian denoising function proposed in
  "Semi-Supervised Learning with Ladder Networks"
  """

  def __init__(self, activation=tf.nn.sigmoid, **kwargs):
    super(GaussianDenoising, self).__init__(**kwargs)
    self.activation = K.linear if activation is None else activation

  def _initialize(self):
    input_shape = self.input_shape
    shape = input_shape[1:]
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a1', roles=Weight)
    # self.config.create_params(
    #     constant(1.), shape=shape, name='a2', roles=Weight)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a3', roles=Bias)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a4', roles=Weight)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a5', roles=Bias)

    # self.config.create_params(
    #     constant(0.), shape=shape, name='a6', roles=Weight)
    # self.config.create_params(
    #     constant(1.), shape=shape, name='a7', roles=Bias)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a8', roles=Bias)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a9', roles=Weight)
    # self.config.create_params(
    #     constant(0.), shape=shape, name='a10', roles=Bias)

  def _apply(self, u, mean, std, z_corr):
    mu = self.a1 * self.activation(self.a2 * u + self.a3) + self.a4 * u + self.a5
    v = self.a6 * self.activation(self.a7 * u + self.a8) + self.a9 * u + self.a10

    z_est = (z_corr - mu) * v + mu
    z_est_bn = (z_est - mean) / K.square(std)
    return z_est_bn

  def _transpose(self):
    raise NotImplementedError



from __future__ import print_function, division, absolute_import

import os
os.environ['ODIN'] = 'gpu,float32'
import shutil
from collections import defaultdict

import numpy as np
import tensorflow as tf

from odin import fuel as F
from odin import nnet as N, backend as K
from odin.utils import (ctext, mpi, Progbar, catch_warnings_ignore, stdio,
                        get_logpath)

from sklearn.metrics import accuracy_score, log_loss, f1_score

from helpers import (FEATURE_RECIPE, FEATURE_NAME, PATH_ACOUSTIC_FEATURES,
                     MINIMUM_UTT_DURATION, ANALYSIS_DIR, EXP_DIR,
                     filter_utterances, prepare_dnn_data)

MODEL_ID = 'xvec_mfccmusanrirs.mfcc.5_pad_5_8.fisher_swb_voxceleb1_voxceleb2'
MODEL_ID = 'xvec_mfccmusanrirs.mfcc.5_pad_5_8.fisher_voxceleb1_voxceleb2'
MODEL_ID = 'xvec_mfccmusanrirs.mfcc.5_pad_5_8.fisher_sre10_swb_voxceleb1_voxceleb2'

info = MODEL_ID.split('.')
feat_name = info[1]
utt_length, seq_mode, min_dur, min_utt = info[2].split('_')
exclude_datasets = info[-1].split('_')
# ====== base dir ====== #
BASE_DIR = os.path.join(EXP_DIR, MODEL_ID)
assert FEATURE_RECIPE.replace('_', '') in os.path.basename(BASE_DIR)
assert FEATURE_NAME in os.path.basename(BASE_DIR)
# ====== get the last model ====== #
all_model = sorted([name
                    for name in os.listdir(BASE_DIR)
                    if 'model.ai.' in name],
                   key=lambda x: int(x.split('.')[-1]))
assert len(all_model) > 0, "Cannot find any model.ai. at path: %s" % BASE_DIR
MODEL = os.path.join(BASE_DIR, all_model[-1])
# ====== prepare log ====== #
stdio(get_logpath(name="analyze.log", increasing=True,
                  odin_base=False, root=ANALYSIS_DIR))
print(ctext(BASE_DIR, 'lightyellow'))
print(ctext(MODEL, 'lightyellow'))
print("Feature name:", ctext(feat_name, 'lightyellow'))
print("Utt length  :", ctext(utt_length, 'lightyellow'))
print("Seq mode    :", ctext(seq_mode, 'lightyellow'))
print("Min Duration:", ctext(min_dur, 'lightyellow'))
print("Min #Utt    :", ctext(min_utt, 'lightyellow'))
print("Excluded    :", ctext(exclude_datasets, 'lightyellow'))
# ===========================================================================
# Load the data
# ===========================================================================
train, valid, all_speakers, ds = prepare_dnn_data(save_dir=BASE_DIR,
    feat_name=feat_name, utt_length=int(utt_length), seq_mode=str(seq_mode),
    min_dur=int(min_dur), min_utt=int(min_utt),
    exclude=exclude_datasets, train_proportion=0.5,
    return_dataset=True)
print(ds)

label2spk = {i: spk for i, spk in enumerate(all_speakers)}
labels = np.arange(len(all_speakers))

X = ds[FEATURE_NAME]
indices = ds['indices_%s' % FEATURE_NAME]
spkid = ds['spkid']
# ===========================================================================
# Load the model
# ===========================================================================
# ====== load the network ====== #
x_vec = N.deserialize(path=MODEL, force_restore_vars=True)
# ====== get output tensors ====== #
y_logit = x_vec()
y_proba = tf.nn.softmax(y_logit)
X = K.ComputationGraph(y_proba).placeholders[0]
z = K.ComputationGraph(y_proba).get(roles=N.Dense, scope='LatentOutput',
                                    beginning_scope=False)[0]
f_prob = K.function(inputs=X, outputs=y_proba, training=False)
f_z = K.function(inputs=X, outputs=z, training=False)
print('Inputs:', ctext(X, 'cyan'))
print('Predic:', ctext(y_proba, 'cyan'))
print('Latent:', ctext(z, 'cyan'))
# ===========================================================================
# Helper
# ===========================================================================
def evaluate_prediction(name_list, y_pred, y_true, title):
  def _report(y_p, y_t, pad=''):
    with catch_warnings_ignore(Warning):
      z_ = np.concatenate(y_p, axis=0)
      z = np.concatenate(y_t, axis=0)
      print(pad, '*** %s ***' % ctext('Frame-level', 'lightcyan'))
      print(pad, "#Samples:", ctext(len(z), 'cyan'))
      print(pad, "Log loss:", log_loss(y_true=z, y_pred=z_, labels=labels))
      print(pad, "Accuracy:", accuracy_score(y_true=z, y_pred=np.argmax(z_, axis=-1)))

      z_ = np.concatenate([np.mean(i, axis=0, keepdims=True) for i in y_p],
                          axis=0)
      z = np.array([i[0] for i in y_t])
      print(pad, '*** %s ***' % ctext('Utterance-level', 'lightcyan'))
      print(pad, "#Samples:", ctext(len(z), 'cyan'))
      print(pad, "Log loss:", log_loss(y_true=z, y_pred=z_, labels=labels))
      print(pad, "Accuracy:", accuracy_score(y_true=z, y_pred=np.argmax(z_, axis=-1)))

  datasets_2_samples = defaultdict(list)
  for name, y_p, y_t in zip(name_list, y_pred, y_true):
    dsname = ds['dsname'][name]
    datasets_2_samples[dsname].append((name, y_p, y_t))

  print('=' * 12, ctext(title, 'lightyellow'), '=' * 12)
  _report(y_p=y_pred, y_t=y_true)

  for dsname, data in sorted(datasets_2_samples.items(),
                             key=lambda x: x[0]):
    print(ctext(dsname, 'yellow'), ':')
    y_pred = [i[1] for i in data]
    y_true = [i[2] for i in data]
    _report(y_p=y_pred, y_t=y_true, pad='  ')
# ===========================================================================
# make prediction
# ===========================================================================
def make_prediction(feeder, title):
  prog = Progbar(target=len(feeder), print_summary=True, name=title)
  name_list = []
  y_pred = []
  y_true = []
  for name, idx, X, y in feeder.set_batch(batch_size=100000,
                                          batch_mode='file',
                                          seed=None, shuffle_level=0):
    name_list.append(name)

    y = np.argmax(y, axis=-1)
    assert len(np.unique(y)) == 1, name
    spk = label2spk[y[0]]
    assert spkid[name] == spk, name
    y_true.append(y)

    y_ = f_prob(X)
    y_pred.append(y_)

    assert len(y) == len(y_)
    prog.add(X.shape[0])
  evaluate_prediction(name_list, y_pred, y_true, title=title)
# ====== do it ====== #
make_prediction(train, title="Train Data")
make_prediction(valid, title="Valid Data")



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
import numpy as np

from odin import preprocessing as pp
from odin import fuel as F, visual as V
from odin.utils import (ctext, get_logpath, get_module_from_path,
                        get_script_path, Progbar, mpi,
                        catch_warnings_ignore)

from helpers import (ALL_FILES, ALL_NOISE, ALL_DATASET, IS_DEBUGGING,
                     PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE,
                     AUGMENTATION_NAME, Config,
                     EXP_DIR, NCPU, validate_features_dataset)

if AUGMENTATION_NAME == 'None':
  raise ValueError("`-aug` option was not provided, choose: 'rirs' or 'musan'")
np.random.seed(Config.SUPER_SEED)
# percentage of data will be used for augmentation
PERCENTAGE_AUGMENTATION = 0.8
# ===========================================================================
# Constant
# ===========================================================================
AUGMENTATION_DATASET = [
    'swb', 'sre04', 'sre05', 'sre06', 'sre08', 'sre10']
AUGMENTATION_DATASET = [i for i in AUGMENTATION_DATASET
                        if i in ALL_DATASET]
print("Augmenting following dataset: %s" %
  ctext(', '.join(AUGMENTATION_DATASET), 'yellow'))
# ====== get the duration ====== #
path = os.path.join(PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE)
assert os.path.exists(path), \
"Acoustic feature must be extracted first, and stored at path: %s" % path
ds = F.Dataset(path, read_only=True)
all_duration = dict(ds['duration'].items())
ds.close()
# ====== select a new file list ====== #
AUG_FILES = []
missing_duration = []
for row in ALL_FILES:
  if row[4] not in AUGMENTATION_DATASET:
    continue
  if row[2] not in all_duration:
    missing_duration.append(row)
    continue
  dur = all_duration[row[2]]
  AUG_FILES.append([i for i in row] + [dur])
print("#Files missing duration:", ctext(len(missing_duration), 'cyan'))
assert len(AUG_FILES), "Cannot find any files for augmentation"
AUG_FILES = np.array(AUG_FILES)
org_shape = AUG_FILES.shape
# select half of the files
np.random.shuffle(AUG_FILES); np.random.shuffle(AUG_FILES)
AUG_FILES = AUG_FILES[:int(len(AUG_FILES) * PERCENTAGE_AUGMENTATION)]
sel_shape = AUG_FILES.shape
# print the log
print("#Augmentation Files:")
print("  Found    :", ctext(org_shape, 'cyan'))
print("  Selected :", ctext(sel_shape, 'cyan'))
# ===========================================================================
# Get the recipe
# ===========================================================================
recipe = get_module_from_path(identifier=FEATURE_RECIPE,
                              path=get_script_path(),
                              prefix='feature_recipes')
assert len(recipe) > 0, "Cannot find recipe with name: %s" % recipe
recipe = recipe[0](augmentation=AUGMENTATION_NAME)
# ===========================================================================
# Debugging
# ===========================================================================
if IS_DEBUGGING:
  with catch_warnings_ignore(Warning):
    n_samples = 120
    prog = Progbar(target=n_samples, print_summary=True,
                   name='Debugging Augmentation')
    for feat in mpi.MPI(jobs=AUG_FILES[:n_samples],
                        func=recipe.transform,
                        ncpu=NCPU, batch=1):
      if np.random.rand() > 0.8:
        feat = {i: j[:1200] if isinstance(j, np.ndarray) else j
                for i, j in feat.items()}
        V.plot_multiple_features(feat, fig_width=20, title=feat['name'])
      prog['name'] = feat['name'][:48]
      prog['dsname'] = feat['dsname']
      prog['dsnoise'] = feat['dsnoise']
      prog.add(1)
    V.plot_save(os.path.join(EXP_DIR, 'debug_%s_%s.pdf' %
                            (FEATURE_RECIPE, AUGMENTATION_NAME)))
    # ====== save the extractor debugging log ====== #
    pp.set_extractor_debug(recipe, debug=True)
    recipe.transform(AUG_FILES[0])
    with open(os.path.join(EXP_DIR, 'debug_%s_%s.log' %
                           (FEATURE_RECIPE, AUGMENTATION_NAME)), 'w') as f:
      for name, step in recipe.steps:
        f.write(step.last_debugging_text)
  exit()
# ===========================================================================
# Run the processor
# ===========================================================================
# ====== basic path ====== #
output_dataset_path = os.path.join(PATH_ACOUSTIC_FEATURES,
                                   '%s_%s' % (FEATURE_RECIPE, AUGMENTATION_NAME))
processor_log_path = os.path.join(EXP_DIR,
  'processor_%s_%s.log' % (FEATURE_RECIPE, AUGMENTATION_NAME))
ds_validation_path = os.path.join(EXP_DIR,
  'validate_%s_%s.pdf' % (FEATURE_RECIPE, AUGMENTATION_NAME))
# ====== running the processing ====== #
with catch_warnings_ignore(Warning):
  # ====== start processing ====== #
  processor = pp.FeatureProcessor(
      jobs=AUG_FILES,
      path=output_dataset_path,
      extractor=recipe,
      n_cache=320,
      ncpu=NCPU,
      override=True,
      identifier='name',
      log_path=processor_log_path,
      stop_on_failure=False)
  processor.run()
# ===========================================================================
# Validation
# ===========================================================================
validate_features_dataset(output_dataset_path, ds_validation_path)



from __future__ import print_function, division, absolute_import

import os
import inspect
from collections import Mapping
from contextlib import contextmanager
from six.moves import cPickle, builtins

from odin.utils import is_string, is_path, as_tuple
from odin.autoconfig import (auto_config, get_floatX, get_session,
                         get_rng, randint, get_random_seed)
auto_config()
floatX = get_floatX()
# ==================== import utilities modules ==================== #
from odin.backend import keras_helpers
from odin.backend.helpers import *
from odin.backend.tensor import *
from odin.backend.activations import *
from odin.backend import role
from odin.backend import metrics
from odin.backend import losses
from odin.backend import optimizers
from odin.backend import rand
from odin.backend import rnn_cell

# This is a bit of an "evil" approach but tie the ODIN Session and
# keras session right when initialize ODIN
keras_helpers.tied_session()

@contextmanager
def variable_dtype(dtype):
  """Temporary change the default dtype for creating variable using ODIN"""
  global floatX
  _old_floatX = str(floatX)
  floatX = dtype
  yield
  floatX = _old_floatX

def initialize_all_variables(vars=None):
  """ This function will automatically check if the variables
  are initialized, and only perform initialization for
  un-initialized variables.

  Note
  ----
  Re-initialize an initialized variable will give it random values
  """
  if vars is None:
    vars = get_all_variables()
  else:
    vars = [v for v in as_tuple(vars)
            if is_variable(v)]
  # ====== check if variable not initialized ====== #
  init_info = eval([tf.is_variable_initialized(v) for v in vars])
  vars = [v for v, inited in zip(vars, init_info)
          if not inited]
  # ====== build mapping graph -> list of vars ====== #
  graph = defaultdict(list)
  for v in vars:
    graph[v.graph].append(v)
  # ====== run the initialization ====== #
  for g, v in graph.items():
    get_session(graph=g).run([i.initializer for i in v])


def is_variable_initialized(var):
  if not is_variable(var):
    raise ValueError("`var` must be instance of tensorflow.Variable, "
                     "but given type: %s" % type(var))
  return eval(tf.is_variable_initialized(var))


def placeholder(shape=None, dtype=floatX, name=None, roles=[]):
  if shape is None and name is None:
    raise ValueError("shape and name arguments cannot be None at the same time.")
  # ====== check duplicated placeholder ====== #
  if name is not None:
    all_placeholders = [
        o._outputs[0] for o in get_all_operations(otype='Placeholder')]
    for v in all_placeholders:
      v_shape = tuple(v.shape.as_list())
      if v.name == name + ':0': # found duplicated variable
        # set new value for variable
        if shape is not None:
          if v_shape == shape:
            return role.add_roles(v, roles)
          else:
            raise ValueError("Pre-defined placeholder with name: %s and"
                " shape: %s, which is different from given shape: %s"
                % (name, v_shape, shape))
        # just get the variable
        else:
          return role.add_roles(v, roles)
  # ====== Modify add name prefix ====== #
  plh = tf.placeholder(dtype=dtype, shape=shape, name=name)
  return role.add_roles(plh, roles)

# ===========================================================================
# Fast evaluation
# ===========================================================================
def _validate_updates(updates):
  if updates is not None:
    if isinstance(updates, Mapping):
      updates = updates.items()
    # create updates ops
    if not isinstance(updates, tf.Operation):
      updates_ops = []
      for u in updates:
        if isinstance(u, (tuple, list)):
          p, new_p = u
          updates_ops.append(tf.assign(p, new_p))
        elif is_operation(u): # assumed already an assign op
          updates_ops.append(u)
      updates = tf.group(*updates_ops)
  return updates

def eval(x, feed_dict=None,
         update_before=None, update_after=None,
         options=None, run_metadata=None):
  ''' Generalized version of code evaluation, it
  could evaluate python and tensorflow expression.

  Parameters
  ----------
  x : list, tuple, dictionary, `Tensor`
      tensorfow `Tensor` for evaluation
  feed_dict : dict
      Input dictionary, mapping placeholder -> values
  update_before: {None, list, or dict}
      mapping from `Tensor` to its new value which is `Tensor` or
      real value, the updates is runned before evaluating
  update_after: {None, list, or dict}
      same as `updates_before`, but run the `updates` after
      evaluate `x`
  options: tensorflow.RunOptions
      thhe options allow controlling the behavior of
      this particular step (e.g. turning tracing on).
  run_metadata: tensorflow.RunMetadata
      When appropriate, the non-Tensor output of this
      step will be collected there. For example,
      when users turn on tracing in options, the
      profiled info will be collected into
      this argument and passed back.

  Example
  -------
  >>> import tensorflow as tf
  >>> from odin import backend as K
  >>> run_metadata = tf.RunMetadata()
  >>> K.eval(...,
  ...        options=tf.RunOptions(trace_level=tf.RunOptions.FULL_TRACE,
  ...                              output_partition_graphs=True),
  ...        run_metadata=run_metadata)
  >>> with open('log_path', 'w') as f:
  >>>   f.write(str(run_metadata))

  Note
  ----
  If "Couldn't open CUDA library libcupti.so.8.0" appears when you
  adding RunOptions, try adding "/usr/local/cuda/extras/CUPTI/lib64/"
  to your LD_LIBRARY_PATH
  '''
  results = ()
  update_before = _validate_updates(update_before)
  update_after = _validate_updates(update_after)
  # ====== run updates before ====== #
  if update_before is not None:
    get_session(update_before.graph).run(update_before, feed_dict=feed_dict,
                                   options=options,
                                   run_metadata=run_metadata)
  # ====== list of Tensor or string ====== #
  if isinstance(x, (tuple, list)):
    string_eval = []
    tensor_eval = []
    tensor_idx = []
    # evaluate string expression
    for i, j in enumerate(x):
      if is_string(j):
        string_eval.append(builtins.eval(j))
      else:
        tensor_eval.append(j)
        tensor_idx.append(i)
    # evaluate tensor
    if len(tensor_eval) > 0:
      graph = [i.graph for i in tensor_eval]
      if len(set(graph)) > 1:
        raise RuntimeError("Cannot evaluate multiple `Tensor` come from "
                           "different `Graph`.")
      tensor_eval = get_session(graph[0]).run(tensor_eval,
                                              feed_dict=feed_dict,
                                              options=options,
                                              run_metadata=run_metadata)
    results = tuple([tensor_eval.pop(0) if i in tensor_idx else string_eval.pop(0)
                     for i in range(len(x))])
  # ====== mapping ====== #
  elif isinstance(x, Mapping):
    results = {}
    tensor_eval_key = []
    tensor_eval_value = []
    for k, v in x.items():
      if is_string(v):
        results[k] = builtins.eval(v)
      else:
        tensor_eval_key.append(k)
        tensor_eval_value.append(v)
    # evaluate tensor
    if len(tensor_eval) > 0:
      graph = [i.graph for i in tensor_eval_value]
      if len(set(graph)) > 1:
        raise RuntimeError("Cannot evaluate multiple `Tensor` come from "
                           "different `Graph`.")
      tensor_eval_value = get_session(graph[0]).run(tensor_eval_value,
                                                    feed_dict=feed_dict,
                                                    options=options,
                                                    run_metadata=run_metadata)
    # update results
    for k, v in zip(tensor_eval_key, tensor_eval_value):
      results[k] = v
  # ====== just a string ====== #
  elif is_string(x):
    results = builtins.eval(x)
  # ====== just a Tensorflow object ====== #
  elif isinstance(x, tf.Operation) or \
  is_tensor(x, inc_distribution=True, inc_variable=True):
    results = get_session(x.graph).run(x, feed_dict=feed_dict,
                                       options=options,
                                       run_metadata=run_metadata)
  # ====== exception ====== #
  else:
    raise RuntimeError("Cannot evaluate object of type: %s" % type(x))
  # ====== run updates after ====== #
  if update_after is not None:
    get_session(update_after.graph).run(update_after, feed_dict=feed_dict,
                                        options=options,
                                        run_metadata=run_metadata)
  return results

# ===========================================================================
# IO helper
# ===========================================================================
_saver = {}

def save_variables(var_list, path, session=None):
  """ This function only apply for trainable parameters """
  if session is None:
    session = get_session()
  var_list = [v for v in set(as_tuple(var_list)) if is_variable(v)]
  name = '|'.join(sorted([v.name for v in var_list]))
  if name in _saver:
    saver = _saver[name]
  else:
    saver = tf.train.Saver(var_list=var_list, restore_sequentially=False,
        allow_empty=False)
  # ====== save the variables ====== #
  checkpoint = saver.save(session, path, global_step=None,
      write_meta_graph=False, write_state=False)
  # ====== save meta-info for recreate variable ====== #
  var_meta = []
  for v in var_list:
    name = v.name.split(':')[0]
    dtype = v.dtype.base_dtype.name
    shape = v.shape.as_list()
    var_meta.append((name, dtype, shape))
  # ====== save the collections ====== #
  collections = {var.name: role.get_roles(var, return_string=True)
                 for var in var_list}
  with open(path + '.collections', 'wb') as f:
    cPickle.dump([collections, var_meta], f,
                 protocol=cPickle.HIGHEST_PROTOCOL)
  return checkpoint

def restore_variables(path, session=None):
  if session is None:
    session = get_session()
  # ====== load and check var meta ====== #
  with open(path + '.collections', 'rb') as f:
    collections, var_meta = cPickle.load(f)
  var_list = []
  allvars = {v.name.split(':')[0]: v for v in get_all_variables()}
  for name, dtype, shape in var_meta:
    if name in allvars: # found predefined variable
      var_list.append(allvars[name])
    else: # create new variable
      if tf.get_variable_scope().name:
        raise RuntimeError("The current variable scope is: %s, you can "
            "only restore variables from default scope."
            % tf.get_variable_scope().name)
      var_list.append(tf.get_variable(
          shape=shape, name=name, dtype=dtype))
  # ====== restore the variables ====== #
  name = '|'.join(sorted([v.name for v in var_list]))
  if name in _saver:
    saver = _saver[name]
  else:
    saver = tf.train.Saver(var_list=var_list, restore_sequentially=False,
                           allow_empty=False)
  saver.restore(session, path)
  # ====== restore the collections ====== #
  for v in var_list:
    role.add_roles(v, collections[v.name])

def save_graph(path, graph=None):
  g = tf.summary.FileWriter(path)
  if graph is None:
    graph = get_session().graph
  elif isinstance(graph, tf.Session):
    graph = graph.graph
  g.add_graph(graph)
  g.flush()
  g.close()



from __future__ import print_function, division, absolute_import

import warnings
from collections import Counter, Mapping

import numpy as np

from odin.fuel.recipe_base import FeederRecipe
from odin.preprocessing.signal import (segment_axis, stack_frames,
                                       mvn as _mvn, wmvn as _wmvn)
from odin.utils import (axis_normalize, is_pickleable, as_tuple, is_number,
                        is_string)

# ===========================================================================
# Helper
# ===========================================================================
def _get_data_label_idx(data_idx, label_idx, ndim):
  data_idx = axis_normalize(axis=data_idx, ndim=ndim, return_tuple=True)
  label_idx = axis_normalize(axis=label_idx, ndim=ndim, return_tuple=True)
  # exclude all index in label_idx
  data_idx = [i for i in data_idx if i not in label_idx]
  return data_idx, label_idx


def _check_label_mode(mode):
  if is_number(mode):
    return np.clip(float(mode), 0., 1.)
  if is_string(mode):
    mode = mode.lower()
    if mode == 'mid':
      mode = 'middle'
    if mode not in ('common', 'last', 'first', 'middle'):
      raise ValueError(
          "`label_mode` can be: 'common', 'last', 'first', 'middle'")
    return mode
  raise ValueError("No support for `label_mode`=%s" % str(mode))


def _apply_label_mode(y, mode):
  # This applying the label transform to 1-st axis
  if is_number(mode):
    n = y.shape[1]
    n = int(float(mode) * n)
    return y[:, n]
  if mode == 'common':
    raise NotImplementedError
  if mode == 'last':
    return y[:, -1]
  elif mode == 'first':
    return y[:, 0]
  elif mode == 'middle':
    n = y.shape[1]
    if n % 2 == 0:
      n //= 2
    else:
      n = n // 2 + 1
    return y[:, n]
  raise NotImplementedError("No support for label mode: '%s'" % mode)


# ===========================================================================
# Shape manipulation
# ===========================================================================
class Indexing(FeederRecipe):

  """ Indexing

  Parameters
  ----------
  idx: any object support `__getitem__`
      pass
  threshold: None, float, call-able
      pass
  mvn: bool
      if True, applying mean-variance normalization
  data_idx: int, list of int
      any Data at given index will be indexed by `idx`
  label_idx: int, list of int
      any Data at given index won't be applied normalization,
      since normalization on integer label can make everything
      go zero!
  """

  def __init__(self, idx, threshold=None,
               mvn=False, varnorm=True,
               data_idx=None, label_idx=()):
    super(Indexing, self).__init__()
    if not hasattr(idx, '__getitem__'):
      raise ValueError("`sad` must has attribute __getitem__ which takes "
          "file name as input and return array of index, same length as data.")
    if threshold is not None and \
    not hasattr(threshold, '__call__') and \
    not is_number(threshold):
      raise ValueError("`threshold` can be None, call-able, or number.")
    self.idx = idx
    self.threshold = threshold
    self.data_idx = data_idx
    self.label_idx = label_idx
    # ====== for normalization ====== #
    self.mvn = bool(mvn)
    self.varnorm = bool(varnorm)

  def _get_index(self, name):
    index = self.idx[name]
    if self.threshold is None:
      index
    elif hasattr(self.threshold, '__call__'):
      index = self.threshold(index)
    elif is_number(self.threshold):
      index = index >= float(self.threshold)
    if index.dtype != np.bool:
      index = index.astype('bool')
    return index

  def process(self, name, X):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(X),
                              return_tuple=True)
    label_idx = axis_normalize(axis=self.label_idx,
                               ndim=len(X),
                               return_tuple=True)
    index = self._get_index(name)
    # ====== indexing ====== #
    X_new = []
    for i, x in enumerate(X):
      if i in data_idx:
        x = x[index]
        # if NOT label, normalization
        if self.mvn and i not in label_idx:
          x = _mvn(x, varnorm=self.varnorm)
      X_new.append(x)
    return name, X_new

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(shapes),
                              return_tuple=True)
    shapes_new = []
    for i, (shp, ids) in enumerate(shapes):
      if i in data_idx:
        ids_new = []
        n_total = 0
        # ====== update the indices ====== #
        for name, _ in ids:
          # this take a lot of time, but
          # we only calculate new shapes once.
          index = self._get_index(name)
          n = np.sum(index)
          n_total += n
          ids_new.append((name, n))
        # ====== update the shape ====== #
        ids = ids_new
        shp = (n_total,) + shp[1:]
      shapes_new.append((shp, ids))
    return shapes_new


class Slice(FeederRecipe):
  """ Slice
  Parameters
  ----------
  indices: int, slice, list of int(or slice)
      for example: [slice(0, 12), slice(20, 38)] will becomes
      x = np.hstack([x[0:12], x[20:38]])
  axis: int
      the axis will be applied given indices
  data_idx: int, list of int, None
      in case Feeders is given multiple Data, target_data is
      the index of Data that will be applied given indices.
      if None is given, the Slice is applied to all Data

  Note
  ----
  Docs for python `slice` as reference: `slice(stop)`,
  `slice(start, stop[, step])`. Create a slice object.
  This is used for extended slicing (e.g. a[0:10:2]).
  """

  def __init__(self, slices, axis, data_idx=None):
    super(Slice, self).__init__()
    # ====== validate axis ====== #
    if not is_number(axis):
      raise ValueError('axis for Slice must be an integer.')
    self.axis = int(axis)
    # ====== validate indices ====== #
    if is_number(slices):
      slices = slice(int(slices), int(slices + 1))
    elif isinstance(slices, (tuple, list)):
      slices = [i if isinstance(i, slice) else slice(int(i), int(i + 1))
                for i in slices
                if isinstance(i, slice) or is_number(i)]
    elif not isinstance(slices, slice):
      raise ValueError('indices must be int, slice, or list of int and slice.')
    self.slices = slices
    # ====== validate target_data ====== #
    self.data_idx = data_idx

  def process(self, name, X):
    X_new = []
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(X),
                              return_tuple=True)
    for _, x in enumerate(X):
      # apply the indices if _ in target_data
      if _ in data_idx:
        ndim = x.ndim
        axis = self.axis % ndim
        # just one index given
        if isinstance(self.slices, (slice, int)):
          indices = tuple([slice(None) if i != axis else self.slices
                           for i in range(ndim)])
          x = x[indices]
        # multiple indices are given
        else:
          indices = []
          for idx in self.slices:
            indices.append(tuple([slice(None) if i != axis else idx
                                  for i in range(ndim)]))
          x = np.concatenate([x[i] for i in indices], axis=self.axis)
        # check if array still contigous
        x = np.ascontiguousarray(x)
      X_new.append(x)
    return name, X_new

  def _from_indices(self, n):
    """ This function estimates number of sample given indices """
    # slice indices
    indices = (self.slices,) if isinstance(self.slices, slice) \
        else self.slices
    count = 0
    for idx in indices:
      idx = idx.indices(n)
      count += idx[1] - idx[0]
    return count

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(shapes),
                              return_tuple=True)
    new_shapes = []
    # ====== check if first dimension is sliced ====== #
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx:
        if self.axis == 0:
          ids = [(name, self._from_indices(length))
                 for name, length in ids]
          n = sum(i[1] for i in ids)
          shp = (n,) + shp[1:]
        else:
          axis = self.axis % len(shp) # axis in case if negative
          # int indices, just 1
          n = self._from_indices(shp[axis])
          shp = tuple([j if i != axis else n
                       for i, j in enumerate(shp)])
      new_shapes.append((shp, ids))
    return new_shapes


class HStack(FeederRecipe):
  """Horizontal Stacking
  merge a list of np.ndarray horizontally (i.e. the last axis)
  into 1 np.array

  Note
  ----
  The new value will be `prepended` to the data list
  """

  def __init__(self, data_idx=None):
    super(HStack, self).__init__()
    self.data_idx = data_idx

  def process(self, name, X):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(X),
                              return_tuple=True)
    if len(X) > 1 and len(data_idx) > 1:
      X_old = [x for i, x in enumerate(X) if i not in data_idx]
      X_new = [x for i, x in enumerate(X) if i in data_idx]
      X = [np.hstack(X_new)] + X_old
    return name, X

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(shapes),
                              return_tuple=True)
    # just 1 shape, nothing to merge
    if len(shapes) <= 1 or len(data_idx) <= 1:
      return shapes
    # merge
    old_shapes = []
    new_shapes = []
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx:
        new_shapes.append((shp, ids))
      else:
        old_shapes.append((shp, ids))
    # ====== horizontal stacking ====== #
    shape, ids = new_shapes[0]
    new_shapes = (
        shape[:-1] + (sum(shp[-1] for shp, _ in new_shapes),),
        ids
    )
    return [new_shapes] + old_shapes


class ExpandDims(FeederRecipe):
  """ ExpandDims """

  def __init__(self, axis, data_idx=0):
    super(ExpandDims, self).__init__()
    self.axis = int(axis)
    self.data_idx = data_idx

  def process(self, name, X):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(X),
                              return_tuple=True)
    X = [np.expand_dims(x, axis=self.axis)
         if i in data_idx else x
         for i, x in enumerate(X)]
    return name, X

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(shapes),
                              return_tuple=True)
    new_shapes = []
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx:
        shp = list(shp)
        axis = self.axis if self.axis >= 0 else \
            (len(shp) + 1 - self.axis)
        shp.insert(axis, 1)
        shp = tuple(shp)
      new_shapes.append((shp, ids))
    return new_shapes


# ===========================================================================
# Feature grouping
# ===========================================================================
class Stacking(FeederRecipe):
  """
  Parameters
  ----------
  left_context: int
      pass
  right_context: int
      pass
  shift: int, None
      if None, shift = right_context
      else amount of frames will be shifted
  keep_length: bool
      if True, padding zeros to begin and end of `X` to
      make the output array has the same length as original
      array.
  data_idx: int, list of int, or None
      list of all Features indices will be applied
  label_mode: string
      'common': most common label in the sequence of label
      'last': last seen label in the sequence
      'first': first seen label in the sequence
      'middle' or 'mid': middle of the sequence.
  label_idx: int, list of int, None, or empty list, tuple
      which data is specified as label will be treated differently
      based on label_mode

  NOTE
  ----
  You must be carefull applying stacking on label before one-hot
  encoded it.
  """

  def __init__(self, left_context=10, right_context=10,
               shift=1, keep_length=False,
               data_idx=None, label_mode='middle', label_idx=()):
    super(Stacking, self).__init__()
    self.left_context = int(left_context)
    self.right_context = int(right_context)
    self.shift = self.left_context if shift is None else int(shift)
    self.data_idx = data_idx
    self.label_mode = _check_label_mode(label_mode)
    self.label_idx = label_idx
    self.keep_length = bool(keep_length)

  @property
  def frame_length(self):
    return self.left_context + 1 + self.right_context

  def process(self, name, X):
    # not enough data points for stacking
    if X[0].shape[0] < self.frame_length:
      return None
    data_idx, label_idx = _get_data_label_idx(
        self.data_idx, self.label_idx, len(X))
    # ====== stacking  ====== #
    X_new = []
    for idx, x in enumerate(X):
      if idx in data_idx:
        if x.ndim == 1:
          x = np.expand_dims(x, axis=-1)
        x = stack_frames(x, frame_length=self.frame_length,
                         step_length=self.shift,
                         keep_length=self.keep_length)
      elif idx in label_idx:
        if not self.keep_length:
          x = segment_axis(x, frame_length=self.frame_length,
                           step_length=self.shift, axis=0,
                           end='cut')
          x = _apply_label_mode(x, self.label_mode)
        else:
          raise NotImplementedError # TODO
      X_new.append(x)
    return name, X_new

  def shape_transform(self, shapes):
    data_idx, label_idx = _get_data_label_idx(
        self.data_idx, self.label_idx, len(shapes))
    # ====== update the shape and indices ====== #
    new_shapes = []
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx or idx in label_idx:
        # calculate new number of samples
        n = 0; ids_new = []
        for name, n_samples in ids:
          n_samples = 1 + (n_samples - self.frame_length) // self.shift
          ids_new.append((name, n_samples))
          n += n_samples
        # for label_idx, number of features kept as original
        if idx in label_idx:
          shp = (n,) + shp[1:]
        # for data_idx, only apply for 2D
        elif idx in data_idx:
          n_features = shp[1] * self.frame_length if len(shp) == 2 \
              else self.frame_length # 1D case
          shp = (n, n_features)
      new_shapes.append((shp, ids_new))
    # ====== do the shape infer ====== #
    return new_shapes

class StackingSequence(FeederRecipe):
  """ Using `stack_frames` method to create data sequence
  """

  def __init__(self, length, data_idx=None):
    super(StackingSequence, self).__init__()
    self.length = int(length)
    self.data_idx = data_idx

  def process(self, name, X):
    data_idx = axis_normalize(axis=self.data_idx, ndim=len(X), return_tuple=True)
    # ====== stacking  ====== #
    X_new = []
    for idx, x in enumerate(X):
      # stack the data
      if idx in data_idx:
        if x.ndim == 1:
          x = np.expand_dims(x, axis=-1)
        feat_shape = x.shape[1:]
        x = stack_frames(x, frame_length=self.length,
                         step_length=1, keep_length=True, make_contigous=True)
        x = np.reshape(x, newshape=(-1, self.length) + feat_shape)
      X_new.append(x)
    return name, X_new

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx, ndim=len(shapes), return_tuple=True)
    # ====== update the shape and indices ====== #
    new_shapes = []
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx:
        n_samples = shp[0]
        shp = (n_samples, self.length) + shp[1:]
      new_shapes.append((shp, ids))
    # ====== do the shape infer ====== #
    return new_shapes

class Sequencing(FeederRecipe):
  """Generate a new array that chops the given array along the given axis
  into overlapping frames.

  This method has been implemented by Anne Archibald,
  as part of the talk box toolkit
  example::

      segment_axis(arange(10), 4, 2)
      array([[0, 1, 2, 3],
         ( [2, 3, 4, 5],
           [4, 5, 6, 7],
           [6, 7, 8, 9]])

  Parameters
  ----------
  frame_length: int
      the length of each frame
  step_length: int
      the number of array elements by which the frames should overlap
  axis: int
      the axis to operate on; if None, act on the flattened array
  end: str
      what to do with the last frame, if the array is not evenly
          divisible into pieces. Options are:
          - 'cut'    Simply discard the extra values
          - 'wrap'   Copy values from the beginning of the array
          - 'pad'    Pad with a constant value
          - 'ignore' Pad any samples with length <= frame_length, otherwise, ignore
          - 'mix'   'cut' when >= frame_length, otherwise, 'pad'
  pad_value: Number
      the value to use for end='pad'
  pad_mode: 'pre', 'post'
      if "pre", padding or wrapping at the beginning of the array.
      if "post", padding or wrapping at the ending of the array.
  data_idx: int, list of int, None, or empty list, tuple
      list of index of all data will be applied

  Return
  ------
  a ndarray
  The array is not copied unless necessary (either because it is unevenly
  strided and being flattened or because end is set to 'pad' or 'wrap').

  Note
  ----
  It is suggested the use end='cut', and choose relevant short
  `frame_length`, using `pad` or `wrap` will significant increase
  amount of memory usage.

  """

  def __init__(self, frame_length=256, step_length=None,
               end='cut', pad_value=0., pad_mode='post',
               data_idx=None):
    super(Sequencing, self).__init__()
    frame_length = int(frame_length)
    step_length = frame_length // 2 if step_length is None else int(step_length)
    if step_length > frame_length:
      raise ValueError("step_length=%d must be smaller than frame_length=%d"
                       % (step_length, frame_length))
    self.frame_length = frame_length
    self.step_length = step_length
    # ====== check mode ====== #
    end = str(end).lower()
    if end not in ('cut', 'pad', 'wrap', 'ignore', 'mix'):
      raise ValueError(
          "`end` mode support included: 'cut', 'pad', 'wrap', 'ignore', 'mix'")
    self.end = end
    self.pad_value = pad_value
    self.pad_mode = str(pad_mode)
    # ====== transform function ====== #
    # specific index
    self.data_idx = data_idx

  def process(self, name, X):
    # ====== not enough data points for sequencing ====== #
    if self.end == 'cut' and \
    any(x.shape[0] < self.frame_length for x in X):
      return None
    if self.end == 'ignore' and \
    any(x.shape[0] > self.frame_length for x in X):
      return None
    end = self.end
    if end == 'ignore':
      end = 'pad'
    # ====== preprocessing data-idx, label-idx ====== #
    data_idx = axis_normalize(axis=self.data_idx, ndim=len(X),
                              return_tuple=True)
    # ====== segments X ====== #
    X_new = []
    for idx, x in enumerate(X):
      ## for data
      if idx in data_idx:
        if end == 'mix':
          x = segment_axis(a=x,
                           frame_length=self.frame_length,
                           step_length=self.step_length, axis=0,
                           end='cut' if x.shape[0] >= self.frame_length else 'pad',
                           pad_value=self.pad_value, pad_mode=self.pad_mode)
        else:
          x = segment_axis(a=x,
                           frame_length=self.frame_length,
                           step_length=self.step_length, axis=0,
                           end=end, pad_value=self.pad_value,
                           pad_mode=self.pad_mode)
      ## for all
      X_new.append(x)
    return name, X_new

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx, ndim=len(shapes),
                              return_tuple=True)
    # ====== update the indices ====== #
    new_shapes = []
    for idx, (shp, ids) in enumerate(shapes):
      if idx in data_idx:
        # transoform the indices
        n = 0; ids_new = []
        for name, n_samples in ids:
          ## MODE = cut
          if self.end == 'cut':
            if n_samples < self.frame_length:
              n_samples = 0
            else:
              n_samples = 1 + np.floor(
              (n_samples - self.frame_length) / self.step_length)
          ## MODE = ignore and pad
          elif self.end == 'ignore':
            if n_samples > self.frame_length:
              n_samples = 0
            else:
              n_samples = 1
          ## MODE = mix
          elif self.end == 'mix':
            if n_samples < self.frame_length:
              n_samples = 1
            else:
              n_samples = 1 + np.floor(
              (n_samples - self.frame_length) / self.step_length)
          ## MODE = pad or wrap
          else:
            if n_samples < self.frame_length:
              n_samples = 1
            else:
              n_samples = 1 + np.ceil(
              (n_samples - self.frame_length) / self.step_length)
          # make sure everything is integer
          n_samples = int(n_samples)
          if n_samples > 0:
            ids_new.append((name, n_samples))
          n += n_samples
        # transform the shape for data
        if idx in data_idx:
          feat_shape = (shp[-1],) if len(shp) >= 2 else ()
          mid_shape = tuple(shp[1:-1])
          shp = (n, self.frame_length,) + mid_shape + feat_shape
      # end
      new_shapes.append((shp, ids))
    return new_shapes



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'float32,gpu'

import scipy.io
import numpy as np

import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin import training
from odin.utils import (args_parse, ctext, Progbar, as_tuple_of_shape,
                        crypto, stdio)
from odin import fuel as F, visual as V, nnet as N, backend as K

from helpers import (get_model_path, prepare_dnn_data,
                     BATCH_SIZE, EPOCH, LEARNING_RATE,
                     SCORE_SYSTEM_ID, GRADIENT_CLIPPING)
# ===========================================================================
# Create data feeder
# ===========================================================================
(EXP_DIR, MODEL_PATH, LOG_PATH) = get_model_path(system_name='xvec')
stdio(LOG_PATH)
# ====== load the data ====== #
(train, valid,
 all_speakers) = prepare_dnn_data(save_dir=EXP_DIR)
n_speakers = len(all_speakers)
# ====== print some log ====== #
print("Training info:")
print('  ', "Batch size       :", ctext(BATCH_SIZE, 'cyan'))
print('  ', "Epoch            :", ctext(EPOCH, 'cyan'))
# ===========================================================================
# Create the network
# ===========================================================================
inputs = [K.placeholder(shape=(None,) + shape[1:],
                        dtype='float32',
                        name='input%d' % i)
          for i, shape in enumerate(as_tuple_of_shape(train.shape))]
X = inputs[0]
y = inputs[1]
print("Inputs:", ctext(inputs, 'cyan'))
# ====== get all saved model if exists ====== #
all_models = [os.path.join(EXP_DIR, i) for i in os.listdir(EXP_DIR)]
all_models = sorted([i for i in all_models if MODEL_PATH + '.' in i])
# ====== the network ====== #
if len(all_models) > 0:
  print("Load model at:", ctext(all_models[SCORE_SYSTEM_ID], 'cyan'))
  x_vec = N.deserialize(path=all_models[SCORE_SYSTEM_ID],
                        force_restore_vars=True)
else:
  with N.args_scope(
      ['TimeDelayedConv', dict(time_pool='none', activation=K.relu)],
      ['Dense', dict(activation=K.linear, b_init=None)],
      ['BatchNorm', dict(activation=K.relu)]
  ):
    x_vec = N.Sequence([
        N.Dropout(level=0.3),

        N.TimeDelayedConv(n_new_features=512, n_time_context=5),
        N.TimeDelayedConv(n_new_features=512, n_time_context=5),
        N.TimeDelayedConv(n_new_features=512, n_time_context=7),

        N.Dense(512), N.BatchNorm(),
        N.Dense(1500), N.BatchNorm(),

        N.StatsPool(axes=1, output_mode='concat'),
        N.Flatten(outdim=2),

        N.Dense(512, name="LatentOutput"), N.BatchNorm(),
        N.Dense(512), N.BatchNorm(),

        N.Dense(n_speakers, activation=K.linear,
                b_init=init_ops.constant_initializer(value=0))
    ], debug=1, name='XNetwork')
# ====== create outputs ====== #
y_logit = x_vec(X)
y_proba = tf.nn.softmax(y_logit)
z = K.ComputationGraph(y_proba).get(roles=N.Dense, scope='LatentOutput',
                                    beginning_scope=False)[0]
print('Latent space:', ctext(z, 'cyan'))
# ====== create loss ====== #
ce = tf.losses.softmax_cross_entropy(onehot_labels=y, logits=y_logit)
acc = K.metrics.categorical_accuracy(y_true=y, y_pred=y_proba)
# ====== params and optimizing ====== #
clipvalue = None if GRADIENT_CLIPPING <= 0 else float(GRADIENT_CLIPPING)
optimizer = K.optimizers.RMSProp(lr=LEARNING_RATE, name='Xrms',
                                 clipnorm=clipvalue, clip_alg='total_norm')
# optimizer = K.optimizers.Adam(lr=LEARNING_RATE, name='XAdam')
# optimizer = K.optimizers.SGD(lr=LEARNING_RATE, momentum=0.5, name='Xmomentum')
print("Optimizer:", ctext(optimizer, 'yellow'))
print('  ', "Learning Rate    :", ctext(LEARNING_RATE, 'cyan'))
print('  ', "Gradient Clipping:", ctext(clipvalue, 'cyan'))
updates = optimizer.minimize(
    loss=ce,
    roles=[K.role.TrainableParameter],
    exclude_roles=[K.role.InitialState],
    verbose=True)
K.initialize_all_variables()
# # ====== Functions ====== #
print('Building training functions ...')
f_train = K.function(inputs, [ce, acc, optimizer.norm],
                     updates=updates,
                     training=True)
print('Building testing functions ...')
f_score = K.function(inputs, [ce, acc],
                    training=False)
# Latent spaces
f_z = K.function(inputs=X, outputs=z,
                 training=False)
# ===========================================================================
# Create trainer
# ===========================================================================
print('Start training ...')
task = training.MainLoop(batch_size=BATCH_SIZE, seed=120825,
                         shuffle_level=2, allow_rollback=False,
                         verbose=4)
task.set_checkpoint(path=MODEL_PATH, obj=x_vec,
                    increasing=True, max_checkpoint=-1)
task.set_callbacks([
    training.NaNDetector(task_name='train', patience=-1),
    training.CheckpointEpoch(task_name='train', epoch_percent=0.5),
    # training.EarlyStopGeneralizationLoss('valid', ce,
    #                                      threshold=5, patience=3)
])
task.set_train_task(func=f_train, data=train,
                    epoch=EPOCH, name='train')
task.set_valid_task(func=f_score, data=valid,
                    freq=training.Timer(percentage=0.5),
                    name='valid')
task.run()



from __future__ import division, absolute_import

import tensorflow as tf

from odin import backend as K
from odin.nnet.base import NNOp

# ===========================================================================
# Flatten
# ===========================================================================
class Flatten(NNOp):
  """ Flatten the array from the right.
  i.e. turn shape=(128,28,28) with outdim=2 into shape=(128, 784)
  """

  def __init__(self, outdim=2, **kwargs):
    super(Flatten, self).__init__(**kwargs)
    self.outdim = int(outdim)

  def _apply(self, X):
    ndims = X.get_shape().ndims
    if ndims is not None:
      if ndims == self.outdim:
        return X
      elif ndims < self.outdim:
        raise RuntimeError("Input shape: %s, cannot be flatten to %d-D"
          % (str(X.get_shape()), self.outdim))
    return K.flatten(X, outdim=self.outdim)

  def _transpose(self):
    return Reshape(shape=self.input_shape)


# ===========================================================================
# REshape
# ===========================================================================
class Reshape(NNOp):
  """ More flexible version of reshape operation, could be used
  for dimension shuffling as well

  Example
  -------
  x.shape = (25, 08, 12)
  reshape(shape=([1], [2], [0]))
  => x.shape = (08, 12, 25)
  """

  def __init__(self, shape, **kwargs):
    super(Reshape, self).__init__(**kwargs)
    self.shape = shape

  def _apply(self, X):
    return K.reshape(X, shape=self.shape)

  def _transpose(self):
    return Reshape(shape=self.input_shape)


class Dimshuffle(NNOp):

  def __init__(self, pattern, **kwargs):
    super(Dimshuffle, self).__init__(**kwargs)
    self.pattern = pattern

  def _apply(self, X):
    return K.dimshuffle(X, pattern=self.pattern)

  def _transpose(self):
    return Reshape(shape=self.input_shape)


class Squeeze(NNOp):

  def __init__(self, axis, **kwargs):
    super(Squeeze, self).__init__(**kwargs)
    self.axis = axis

  def _apply(self, X):
    input_shape = X.shape
    if input_shape[self.axis] != 1:
      raise ValueError('The squeeze axis=%d must be 1, but got %d instead' %
                       (self.axis, input_shape[self.axis]))
    return tf.squeeze(X, axis=self.axis)

  def _transpose(self):
    return InvertSqueeze(self)


# NNTranspose
class InvertSqueeze(NNOp):

  def _apply(self, X):
    ndim = len(self.T.input_shape)
    axis = self.T.axis % ndim
    pattern = ['x' if i == axis
               else (i - 1 if i > axis else i)
               for i in range(ndim)]
    return K.dimshuffle(X, pattern)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import math
import random

from numbers import Number
from itertools import chain
from collections import defaultdict, Iterator, OrderedDict, Mapping

import numpy as np

from odin.autoconfig import get_rng
from odin.maths import interp
from odin.utils import as_tuple, flatten_list, ctext, batching

def prior2weights(prior, exponential=False,
                  min_value=0.1, max_value=None,
                  norm=False):
  """
  Parameters
  ----------
  prior: numpy.ndarray [nb_classes,]
      probabilty values of each classes prior,
      sum of all prior must be equal to 1.
  exponential: bool
  min_value: bool
      minimum value for the class with highest prior
  max_value: bool
      maximum value for the class with smalles prior
  norm: bool
      if True, normalize output weights to sum up to 1.
  """
  # idea is the one with highest prior equal to 1.
  # and all other classes is the ratio to this prior
  prior = np.array(prior).ravel()
  # make sure everything sum to 1 (probability values)
  prior = prior / np.sum(prior)
  zero_ids = [i for i, j in enumerate(prior) if j == 0]
  nonzero_prior = np.array([j for i, j in enumerate(prior) if j != 0])
  prior = 1. / nonzero_prior * np.max(nonzero_prior)
  if exponential:
    prior = sorted([(i, p) for i, p in enumerate(prior)],
                   key=lambda x: x[-1], reverse=False)
    alpha = interp.expIn(n=len(prior), power=10)
    prior = {i: a * p for a, (i, p) in zip(alpha, prior)}
    prior = np.array([prior[i] for i in range(len(prior))]) + 1
  # ====== rescale everything within max_value ====== #
  if min_value is not None and max_value is not None:
    min_value = float(min_value)
    max_value = float(max_value)
    prior = (max_value - min_value) * (prior - np.min(prior)) \
        / (np.max(prior) - np.min(prior)) + min_value
  # ====== normaize by ====== #
  if norm:
    prior = prior / np.sum(prior)
  # ====== set zero indices ====== #
  prior = prior.tolist()
  for i in zero_ids:
    prior.insert(i, 0)
  return np.array(prior)

# ===========================================================================
# Diagnose
# ===========================================================================
def _split_list(x, rng, train=0.6, idfunc=None, inc_test=True):
  # ====== shuffle input ====== #
  if idfunc is not None:
    x_id = defaultdict(list)
    for i in x:
      x_id[idfunc(i)].append(i)
  else:
    x_id = {i: [j] for i, j in enumerate(x)}
  # shuffle ID(s)
  id_list = list(x_id.keys())
  rng.shuffle(id_list)
  # ====== split ====== #
  N = len(id_list)
  if N == 1:
    raise ValueError("Only find 1 sample, cannot split")
  train = int(np.floor(float(train) * N))
  if train >= N:
    raise ValueError("train proportion must larger than 0 and smaller than 1.")
  valid = (N - train) // (2 if inc_test else 1)
  # ====== return splitted ====== #
  rets = (flatten_list((x_id[i] for i in id_list[:train]), level=1),
          flatten_list((x_id[i] for i in id_list[train: train + valid]), level=1))
  if inc_test:
    rets += (flatten_list((x_id[i] for i in id_list[train + valid:]), level=1),)
  else:
    rets += ([],)
  assert sum(len(r) for r in rets) == len(x), \
      "Number of returned data inconsitent from original data, %d != %d" % \
      (sum(len(r) for r in rets), len(x))
  return rets

def train_valid_test_split(x, train=0.6, cluster_func=None, idfunc=None,
                           inc_test=True, seed=None):
  """ Split given list into 3 dataset: for training, validating,
  and testing.

  Parameters
  ----------
  x: list, tuple, numpy.ndarray
      the dataset.
  train: float (0.0 - 1.0)
      proportion used for training, validating and testing will be
      half of the remain.
  cluster_func: None or call-able
      organize data into cluster, then applying the same
      train_valid_test split strategy for each cluster.
  idfunc: None or call-able
      a function transform the task list into unique identity for splitting,
      `idfunc` has to return comparable keys.
  inc_test: bool
      if split a proportion of data for testing also.
  seed: int
      random seed to produce a re-producible results.
  """
  if seed is not None:
    rng = np.random.RandomState(seed)
  else:
    rng = get_rng()
  # ====== check input ====== #
  if isinstance(x, Mapping):
    x = x.items()
  elif isinstance(x, np.ndarray):
    x = x.tolist()
  # ====== check idfunc ====== #
  if not hasattr(idfunc, '__call__'):
    idfunc = None
  # ====== clustering ====== #
  if cluster_func is None:
    cluster_func = lambda x: 8 # lucky number
  if not hasattr(cluster_func, '__call__'):
    raise ValueError("'cluster_func' must be call-able or None.")
  clusters = defaultdict(list)
  for i in x:
    clusters[cluster_func(i)].append(i)
  # ====== applying data split for each cluster separately ====== #
  train_list, valid_list, test_list = [], [], []
  for name, clus in clusters.items():
    _1, _2, _3 = _split_list(clus, rng, train=train, idfunc=idfunc,
                             inc_test=inc_test)
    train_list += _1
    valid_list += _2
    test_list += _3
  # ====== return the results ====== #
  if inc_test:
    return train_list, valid_list, test_list
  return train_list, valid_list

def freqcount(x, key=None, count=1, normalize=False, sort=False,
              pretty_return=False):
  """ x: list, iterable

  Parameters
  ----------
  key: call-able
      extract the key from each item in the list
  count: call-able, int
      extract the count from each item in the list
  normalize: bool
      if normalize, all the values are normalized from 0. to 1. (
      which sum up to 1. in total).
  sort: boolean
      if True, the list will be sorted in ascent order.
  pretty_return: boolean
      if True, return pretty formatted text.

  Return
  ------
  dict: x(obj) -> freq(int)
  if `pretty_return` is `True`, return pretty formatted string.
  """
  freq = defaultdict(int)
  if key is None:
    key = lambda x: x
  if count is None:
    count = 1
  if isinstance(count, Number):
    _ = int(count)
    count = lambda x: _
  for i in x:
    c = count(i)
    i = key(i)
    freq[i] += c
  # always return the same order
  s = float(sum(v for v in freq.values()))
  freq = OrderedDict([(k, freq[k] / s if normalize else freq[k])
                      for k in sorted(freq.keys())])
  # check sort
  if sort:
    freq = OrderedDict(sorted(freq.items(), key=lambda x: x[1]))
  # check pretty return
  if pretty_return:
    s = ''
    for name, value in freq.items():
      s += ' %s: %d\n' % (ctext(name, 'yellow'), value)
    return s
  return freq

# ===========================================================================
# Bayesian
# ===========================================================================
def KL_divergence(P, Q):
  """ KL(P||Q) = ∑_i • p_i • log(p_i/q_i)
  The smaller this number, the better P match Q distribution
  """
  if isinstance(P, Mapping) and isinstance(Q, Mapping):
    keys = sorted(P.keys())
    P = [P[k] for k in keys]
    Q = [Q[k] for k in keys]
  # ====== normalize to probability 0-1 ====== #
  P = np.array(P)
  P = P / np.sum(P, axis=-1)
  Q = np.array(Q)
  Q = Q / np.sum(Q, axis=-1)
  # ====== calcuate the KL-div ====== #
  D = 0
  for pi, qi in zip(P, Q):
    D += pi * np.log(pi / qi)
  return D

# ===========================================================================
# Sampler
# ===========================================================================
def sampling_iter(it, k, p=None, return_iter=True, seed=5218,
                  progress_bar=None):
  """ Reservoir sampling, randomly choosing a sample of k items from a
  list S containing n items, where n is either a very large or unknown number.
  Typically n is large enough that the list doesn't fit into main memory.

  Parameters
  ----------
  it : iteration
    any instance of of iteration (i.e. `hasattr` '__iter__' )
  k : int
    number of sample for sampling
  p : {None, float (0.-1.)} (default: None)
    if `p` is None, perform reservoir sampling that guarantee every
    elements of the iteration is equally selected.
    if `p` is scalar, perform decayed sampling with `p` is the start
    probability (recommended in case `k` does not fit in the memory)
  return_iter : bool (default: True)
    if True, return an iteration of results instead of extracted list
  seed : int (default: 5218)
    random seed for reproducibility
  progress_bar : {None, odin.utils.Progbar.ProgBar}
  """
  k = int(k); assert k > 0
  if p is not None:
    p = float(p); assert 0. < p < 1.
  assert hasattr(it, '__iter__')
  # ====== check progress bar ====== #
  if progress_bar is not None:
    from odin.utils import Progbar
    assert isinstance(progress_bar, Progbar),\
    '`progress_bar` must be instance of odin.utils.progbar.Progbar, but given: %s' \
    % str(type(progress_bar))
  # ====== reservoir sampling ====== #
  if p is None:
    random.seed(seed)
    ret = []
    for i, x in enumerate(it):
      if i < k:
        ret.append(x)
      else:
        # as the iteration move forward,
        # the chance of picking new sample decrease
        r = random.randint(0, i)
        if r < k:
          ret[r] = x
      # update progress bar
      if progress_bar is not None:
        progress_bar.add(x)
    return tuple(ret)

  # ====== simulating the probability decay ====== #
  def _sampling():
    n = 0
    prob = p
    ret = []
    # this is compromise of randomness for speed
    n_buffer = 12000
    rand = np.random.RandomState(seed=seed)
    buffered_random = rand.rand(n_buffer)
    buffered_index = rand.randint(0, k, size=n_buffer, dtype=int)
    for i, x in enumerate(it):
      r = buffered_random[i % n_buffer]
      # initialize the reservoir
      if len(ret) < k:
        ret.append(x)
      # sample selected
      elif r < prob:
        yield x
        n += 1
      # update the reservoir
      else:
        ret[buffered_index[i % n_buffer]] = x
      # check break condition
      if n >= k:
        break
      # update the probability
      prob = 0.8 * prob + 0.2 / (i + 1)
      # update progress bar
      if progress_bar is not None:
        progress_bar.add(x)
    # return the rest of the samples to have enough k sample
    for i in range(k - n):
      yield ret[i]
  return _sampling() if return_iter else list(_sampling)

# ===========================================================================
# Statistics
# ===========================================================================
def sparsity_percentage(x, batch_size=5218):
  n_zeros = 0
  n_total = np.prod(x.shape)
  for start, end in batching(batch_size=batch_size, n=x.shape[0],
                             seed=None):
    y = x[start:end]
    n_nonzeros = np.count_nonzero(y)
    n_zeros += np.prod(y.shape) - n_nonzeros
  return n_zeros / n_total

def logVMR(x, axis=None, logged_values=False):
  """ Calculate the variance to mean ratio (VMR) in non-logspace
  (return answer in log-space)

  VMR (variance-to-mean ratio = index of dispersion) is equal to zero
  in the case of a constant random variable (not dispersed).

  It is equal to one in a Poisson distribution, higher than one in a
  negative binomial distribution (over dispersed) and between zero
  and one in a binomial distribution (under dispersed).

  Reference
  ---------
  https://www.quantshare.com/item-1029-index-of-dispersion-variance-to-mean-ratio-vmr

  """
  if logged_values:
    x = np.expm1(x)
  return np.log1p(np.var(x, axis=axis) / np.mean(x, axis=axis))

# ===========================================================================
# Diagnosis
# ===========================================================================
def classification_diagnose(X, y_true, y_pred,
                            num_samples=8, return_list=False, top_n=None,
                            seed=5218):
  """
  Return
  ------
  OrderedDict: (true_class, pred_class) -> [list of samples from `X`]
  sorted by most frequence to less frequence
  """
  # Mapping from classesID -> [Sample ID]
  # ====== check argument ====== #
  if y_true.ndim == 2:
    y_true = np.argmax(y_true, axis=-1)
  elif y_true.ndim != 1:
    raise ValueError("Only support 1-D or 2-D one-hot `y_true`, "
                     "given `y_true` with shape: %s" % str(y_true.shape))
  if y_pred.ndim == 2:
    y_pred = np.argmax(y_pred, axis=-1)
  elif y_pred.ndim != 1:
    raise ValueError("Only support 1-D or 2-D one-hot `y_pred`, "
                     "given `y_pred` with shape: %s" % str(y_pred.shape))
  assert len(y_true) == len(y_pred) == len(X), "Inconsistent number of samples"
  # ====== initialize ====== #
  rand = np.random.RandomState(seed)
  miss = defaultdict(list)
  # ====== sampling ====== #
  for idx, (true, pred) in enumerate(zip(y_true, y_pred)):
    if true != pred:
      miss[(true, pred)].append(idx)
  # sort by the most freq mistake classes
  outputs = OrderedDict()
  for (true, pred), samples in sorted(miss.items(),
                                      key=lambda x: len(x[-1]),
                                      reverse=True):
    rand.shuffle(samples)
    outputs[(true, pred)] = (X[samples[:num_samples]]
                             if isinstance(X, np.ndarray) else
                             [X[i] for i in samples])
  # select top frequence
  if top_n is not None and isinstance(top_n, Number):
    top_n = int(top_n)
    assert top_n >= 1
    outputs = OrderedDict(list(outputs.items())[:top_n])
  return list(outputs.items()) if return_list else outputs

def classification_report(y_pred, y_true, labels):
  """
  Parameters
  ----------
  pass

  Return
  ------
  Classification report in form of string
  """
  from sklearn.metrics import accuracy_score, classification_report, confusion_matrix
  # ====== validate labels ====== #
  labels = as_tuple(labels)
  target_names = [str(i) for i in labels]
  labels = list(range(0, len(labels)))
  # ====== create report ====== #
  s = ""
  s += "Accuracy: %f\n" % accuracy_score(y_true, y_pred, normalize=True)
  s += "Confusion matrix:\n"
  s += str(confusion_matrix(y_true, y_pred, labels=labels)) + '\n'
  s += "Report:\n"
  s += str(classification_report(y_true, y_pred, labels=labels, digits=3,
                                 target_names=target_names))
  return s

def summary(x, axis=None, shorten=False):
  """ Return string of statistical summary given series `x`
    {#:%s|mi:%s|q1:%s|md:%s|mn:%s|q3:%s|ma:%s|sd:%s}
  """
  if isinstance(x, Iterator):
    x = list(x)
  if isinstance(x, (tuple, list, set)):
    x = np.array(x)
  mean, std = np.mean(x, axis=axis), np.std(x, axis=axis)
  median = np.median(x, axis=axis)
  qu1, qu3 = np.percentile(x, [25, 75], axis=axis)
  min_, max_ = np.min(x, axis=axis), np.max(x, axis=axis)
  s = ""
  if not shorten:
    x = x.ravel()
    samples = ', '.join([str(i)
           for i in np.random.choice(x, size=min(8, len(x)), replace=False).tolist()])
    s += "***** Summary *****\n"
    s += "    Min : %s\n" % str(min_)
    s += "1st Qu. : %s\n" % str(qu1)
    s += " Median : %s\n" % str(median)
    s += "   Mean : %g\n" % mean
    s += "3rd Qu. : %s\n" % str(qu3)
    s += "    Max : %s\n" % str(max_)
    s += "-------------------\n"
    s += "    Std : %g\n" % std
    s += "#Samples: %d\n" % len(x)
    s += "Samples : %s\n" % samples
    s += "Sparsity: %.4f\n" % sparsity_percentage(x)
  else:
    s += "{#:%s|mi:%s|q1:%s|md:%s|mn:%s|q3:%s|ma:%s|sd:%s}" %\
    (ctext(len(x), 'cyan'),
     ctext('%g' % min_, 'cyan'),
     ctext('%g' % qu1, 'cyan'),
     ctext('%g' % median, 'cyan'),
     ctext('%g' % mean, 'cyan'),
     ctext('%g' % qu3, 'cyan'),
     ctext('%g' % max_, 'cyan'),
     ctext('%g' % std, 'cyan'))
  return s

describe = summary



from __future__ import division, absolute_import

from theano import tensor as T
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams
import numpy as np

from odin.autoconfig import get_rng, CONFIG
from odin.backend import tensor as K

FLOATX = CONFIG.floatX
EPSILON = CONFIG.epsilon
PI = np.pi
C = -0.5 * np.log(2 * PI)
_RNG = RandomStreams(seed=get_rng().randint(10e8))

# ===========================================================================
# Variational OPERATIONS
# ===========================================================================
def log_prob_bernoulli(p_true, p_approx, mask=None):
  """ Compute log probability of some binary variables with probabilities
  given by p_true, for probability estimates given by p_approx. We'll
  compute joint log probabilities over row-wise groups.
  Note
  ----
  origin implementation from:
  https://github.com/Philip-Bachman/ICML-2015/blob/master/LogPDFs.py
  Copyright (c) Philip Bachman
  """
  if mask is None:
    mask = T.ones((1, p_approx.shape[1]))
  log_prob_1 = p_true * T.log(p_approx)
  log_prob_0 = (1.0 - p_true) * T.log(1.0 - p_approx)
  log_prob_01 = log_prob_1 + log_prob_0
  row_log_probs = T.sum((log_prob_01 * mask), axis=1, keepdims=True)
  return row_log_probs

#logpxz = -0.5*np.log(2 * np.pi) - log_sigma_decoder - (0.5 * ((x - mu_decoder) / T.exp(log_sigma_decoder))**2)


def log_prob_gaussian(mu_true, mu_approx, les_sigmas=1.0, mask=None):
  """
  Compute log probability of some continuous variables with values given
  by mu_true, w.r.t. gaussian distributions with means given by mu_approx
  and standard deviations given by les_sigmas.
  Note
  ----
  origin implementation from:
  https://github.com/Philip-Bachman/ICML-2015/blob/master/LogPDFs.py
  Copyright (c) Philip Bachman
  """
  if mask is None:
    mask = T.ones((1, mu_approx.shape[1]))
  ind_log_probs = C - T.log(T.abs_(les_sigmas)) - \
  ((mu_true - mu_approx)**2.0 / (2.0 * les_sigmas**2.0))
  row_log_probs = T.sum((ind_log_probs * mask), axis=1, keepdims=True)
  return row_log_probs


def log_prob_gaussian2(mu_true, mu_approx, log_vars=1.0, mask=None):
  """
  Compute log probability of some continuous variables with values given
  by mu_true, w.r.t. gaussian distributions with means given by mu_approx
  and log variances given by les_logvars.
  Note
  ----
  origin implementation from:
  https://github.com/Philip-Bachman/ICML-2015/blob/master/LogPDFs.py
  Copyright (c) Philip Bachman
  """
  if mask is None:
    mask = T.ones((1, mu_approx.shape[1]))
  ind_log_probs = C - (0.5 * log_vars) - \
  ((mu_true - mu_approx)**2.0 / (2.0 * T.exp(log_vars)))
  row_log_probs = T.sum((ind_log_probs * mask), axis=1, keepdims=True)
  return row_log_probs



from __future__ import division, absolute_import

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin import backend as K
from odin.nnet.base import NNOp
from odin.autoconfig import randint
from odin.utils import as_tuple, is_string
from odin.backend.role import ConvKernel, Bias

# ===========================================================================
# Helper
# ===========================================================================
def __get_deconv_shape_1axis(n, kernel_shape, border_mode,
                             subsample, dilation=1):
  if None in [n, kernel_shape, border_mode,
              subsample, dilation]:
    return None
  if dilation != 1:
    raise ValueError("Deconvolution with dilation != 1 is not supported.")
  # ====== upsample ====== #
  if subsample != 1:
    n = n * subsample
  # ====== padding ====== #
  if isinstance(border_mode, str):
    border_mode = border_mode.lower()
  if border_mode == "half" or border_mode == "same":
    n = n
  elif border_mode == "full": # M + N - 1
    n += 1 - kernel_shape
  elif border_mode == "valid": # M - N + 1
    n += max(kernel_shape - subsample, 0)
  elif border_mode >= 0:
    n += max(kernel_shape - subsample, 0) - border_mode
  else:
    raise ValueError("border_mode must be >= 0")
  return n


def get_deconv_output_shape(image_shape, kernel_shape,
                            border_mode, subsample,
                            filter_dilation=None):
  """ Invert the process of calculating output shape for convolution
  (Deconvolution, or TransposedConvolution)

  Parameters
  ----------
  image_shape: tuple of int (symbolic or numeric) corresponding to the input
      order: (samples, conv_dim1, conv_dim2, conv_dim3, ..., input_depth)
      (i.e tensorflow-NHWC format)
  kernel_shape: tuple of int (symbolic or numeric) corresponding to the
      order: (kernel_dim1, kernel_dim2, kernel_dim3, ..., input_depth, out_depth)
      (i.e tensorflow-NHWC format)
  border_mode: string, int (symbolic or numeric) or tuple of int (symbolic
      or numeric). If it is a string, it must be 'valid', 'half' or 'full'.
      If it is a tuple, its two (or three) elements respectively correspond
      to the padding on height and width (and possibly depth) axis.
  subsample: tuple of int (symbolic or numeric). Its or three elements
      espectively correspond to the subsampling on height and width (and
      possibly depth) axis.
  filter_dilation: tuple of int (symbolic or numeric). Its two elements
      correspond respectively to the dilation on height and width axis.

  Returns
  -------
  output_shape: tuple of int corresponding to the output image shape. Its
      four element must correspond respectively to: batch size, number of
      output channels, height and width of the image. None where undefined.

  """
  # ======  convert tensorflow shape to theano shape ====== #
  image_shape = (image_shape[0], image_shape[-1]) + tuple(image_shape[1:-1])
  kernel_shape = (kernel_shape[-1], kernel_shape[-2]) + tuple(kernel_shape[:-2])
  # ====== infer shape ====== #
  bsize, imshp = image_shape[0], image_shape[2:]
  outkern, kshp = kernel_shape[1], kernel_shape[2:]
  if filter_dilation is None:
    filter_dilation = np.ones(len(subsample), dtype='int')
  if isinstance(border_mode, tuple):
    out_shp = tuple(__get_deconv_shape_1axis(
        imshp[i], kshp[i], border_mode[i],
        subsample[i], filter_dilation[i]) for i in range(len(subsample)))
  else:
    out_shp = tuple(__get_deconv_shape_1axis(
        imshp[i], kshp[i], border_mode,
        subsample[i], filter_dilation[i]) for i in range(len(subsample)))
  # ====== convert theano to tensorflow shape ====== #
  return (bsize, ) + out_shp + (outkern,)


# ===========================================================================
# Ops
# ===========================================================================
class Conv(NNOp):
  """ Convolutional Operator

  Performs a 2D or 3D convolution on its input and optionally adds a bias and
  applies an elementwise activation.

  Parameters
  ----------
  num_filters : int
      The number of learnable convolutional filters this layer has.

  filter_size : int or iterable of int
      tuple specifying the size of the filters.

  strides : int or iterable of int
      specifying the stride of the convolution operation.

  pad : int, iterable of int, 'full', 'same' or 'valid' (default: 'valid')
      By default, the convolution is only computed where the input and the
      filter fully overlap (a valid convolution). When ``stride=1``, this
      yields an output that is smaller than the input by ``filter_size - 1``.
      The `pad` argument allows you to implicitly pad the input with zeros,
      extending the output size.

      A single integer results in symmetric zero-padding of the given size on
      all borders, a tuple of two integers allows different symmetric padding
      per dimension.

      ``'full'`` pads with one less than the filter size on both sides. This
      is equivalent to computing the convolution wherever the input and the
      filter overlap by at least one position.

      ``'same'`` pads with half the filter size (rounded down) on both sides.
      When ``stride=1`` this results in an output size equal to the input
      size. Even filter size is not supported.
                  pad|                                      |pad
      inputs:      0 |1  2  3  4  5  6  7  8  9  10 11 12 13|0  0
                  |________________|
                                 |_________________|
                                                |________________|
      ``'valid'`` is an alias for ``0`` (no padding / a valid convolution).
      inputs:         1  2  3  4  5  6  7  8  9  10 11 (12 13)
                     |________________|                dropped
                                    |_________________|
      Note that ``'full'`` and ``'same'`` can be faster than equivalent
      integer values due to optimizations by theano or tensorflow.

  untie_biases : bool (default: False)
      If ``False``, the layer will have a bias parameter for each channel,
      which is shared across all positions in this channel. As a result, the
      `b` attribute will be a vector (1D).

      If True, the layer will have separate bias parameters for each
      position in each channel. As a result, the `b` attribute will be a
      3D tensor.

  W_init : Theano shared variable, expression, numpy array or call-able
      Initial value, expression or initializer for the weights.
      These should be a 4D tensor with shape
      ``(num_filters, num_input_channels, filter_rows, filter_columns)``.
      See :func:`lasagne.utils.create_param` for more information.

  b_init : Theano shared variable, expression, numpy array, call-able or ``None``
      Initial value, expression or initializer for the biases. If set to
      ``None``, the layer will have no biases. Otherwise, biases should be
      a 1D array with shape ``(num_filters,)`` if `untie_biases` is set to
      ``False``. If it is set to ``True``, its shape should be
      ``(num_filters, output_rows, output_columns)`` instead.
      See :func:`lasagne.utils.create_param` for more information.

  activation : call-able or None
      The activation that is applied to the layer activations. If None
      is provided, the layer will be linear.

  dilation : int or iterable of int
      Specifying the dilation factor of the filters. A factor of
      :math:`x` corresponds to :math:`x - 1` zeros inserted between
      adjacent filter elements.

  **kwargs
      Any additional keyword arguments are passed to the `NNOp` superclass.

  Note
  ----
  This Ops can be used for 1D, 2D (images) and 3D (videos).
  dim_ordering : tf-tensorflow (defaults)
      input shape: (samples, conv_dim1, conv_dim2, input_depth)
      kernel shape: (kernel_dim1, kernel_dim2, input_depth, out_depth)
  Only support float32 on CPU
  """

  def __init__(self, num_filters, filter_size, strides=1, pad='valid',
               W_init=init_ops.glorot_uniform_initializer(seed=randint()),
               b_init=init_ops.constant_initializer(0),
               untie_biases=False, activation=K.linear,
               dilation=1, **kwargs):
    super(Conv, self).__init__(**kwargs)
    self.num_filters = int(num_filters)
    self.filter_size = filter_size
    self.strides = strides
    self.pad = pad
    self.W_init = W_init
    self.b_init = b_init
    self.untie_biases = bool(untie_biases)
    self.dilation = dilation
    self.activation = K.linear if activation is None else activation

  # ==================== abstract methods ==================== #
  @property
  def kernel_shape(self):
    # TF kernel shape: (kernel_dim1, kernel_dim2, ..., input_depth, out_depth)
    return self.filter_size + (self.input_shape[-1], self.num_filters)

  def _transpose(self):
    return TransposeConv(num_filters=self.input_shape[-1],
                         filter_size=self.filter_size,
                         strides=self.strides, pad=self.pad,
                         W_init=self.get('W'),
                         b_init=None if self.b_init is None else 0.,
                         untie_biases=self.untie_biases,
                         activation=self.activation,
                         dilation=self.dilation,
                         output_shape=self.input_shape)

  def _initialize(self):
    # ====== validate init arguments ====== #
    self.ndim = len(self.input_shape) - 2
    # padding
    if isinstance(self.pad, (tuple, list, int)):
      self.pad = as_tuple(self.pad, self.ndim, int)
    elif self.pad is None:
      self.pad = (0,) * self.ndim
    elif is_string(self.pad):
      self.pad = self.pad.upper()
    # strides
    if self.strides is None:
      self.strides = (1,) * self.ndim
    else:
      self.strides = as_tuple(self.strides, self.ndim, int)
    # dilation
    if self.dilation is None:
      self.dilation = (1,) * self.ndim
    else:
      self.dilation = as_tuple(self.dilation, self.ndim, int)
    # filter size
    self.filter_size = as_tuple(self.filter_size, self.ndim, int)
    # ====== create config ====== #
    # weights
    self.get_variable_nnop(initializer=self.W_init,
        shape=self.kernel_shape, name='W', roles=ConvKernel)
    if self.b_init is not None:
      if self.untie_biases:
        biases_shape = self.output_shape[1:]
      else:
        biases_shape = (self.num_filters,)
      self.get_variable_nnop(initializer=self.b_init,
          shape=biases_shape, name='b', roles=Bias)

  def _apply(self, X):
    # ====== apply convolution ====== #
    conved = self.convolve(X)
    # ====== apply bias ====== #
    if 'b' in self.variable_info:
      if self.untie_biases:
        conved += tf.expand_dims(self.get('b'), axis=0)
      else:
        conved += K.dimshuffle(self.get('b'), ('x',) * (self.ndim + 1) + (0,))
    return self.activation(conved)

  def convolve(self, X):
    if self.ndim == 1:
      data_format = "NWC"
    elif self.ndim == 2:
      data_format = "NHWC"
    elif self.ndim == 3:
      data_format = "NDHWC"
    else:
      raise Exception('No support for %d-D input.' % self.ndim)
    # ====== perform normal convolution ====== #
    conved = tf.nn.convolution(input=X, filter=self.get('W'),
        padding=self.pad,
        strides=self.strides,
        dilation_rate=self.dilation,
        data_format=data_format)
    return conved

# ===========================================================================
# TransposeConv
# ===========================================================================
class TransposeConv(Conv):

  def __init__(self, num_filters, filter_size, strides=1, pad='valid',
               W_init=init_ops.glorot_uniform_initializer(seed=randint()),
               b_init=init_ops.constant_initializer(0),
               untie_biases=False, activation=K.linear,
               dilation=1, output_shape=None, **kwargs):
    super(TransposeConv, self).__init__(num_filters=num_filters,
             filter_size=filter_size, strides=strides, pad=pad,
             W_init=W_init, b_init=b_init,
             untie_biases=untie_biases, activation=activation,
             dilation=dilation, **kwargs)
    # explicit output shape
    self._output_shape = output_shape

  def _transpose(self):
    return Conv(num_filters=self.input_shape[-1], filter_size=self.filter_size,
                strides=self.strides, pad=self.pad,
                W_init=self.get('W'),
                b_init=None if self.b_init is None else 0.,
                untie_biases=self.untie_biases,
                activation=self.activation,
                dilation=self.dilation)

  @property
  def kernel_shape(self):
    # TF kernel shape: (kernel_dim1, kernel_dim2, ..., input_depth, out_depth)
    # revert the output_channel in input_channel to keep the same kernel
    # shape as original Convolution ops
    return self.filter_size + (self.num_filters, self.input_shape[-1])

  @property
  def output_shape(self):
    if self._output_shape is None:
      return get_deconv_output_shape(self.input_shape, self.kernel_shape,
              border_mode=self.pad, subsample=self.strides,
              filter_dilation=self.dilation)
    return self._output_shape

  def convolve(self, X):
    if self.ndim == 2:
      deconv_func = tf.nn.conv2d_transpose
    elif self.ndim == 3:
      deconv_func = tf.nn.conv3d_transpose
    else:
      raise Exception('No support for %d-D input.' % self.ndim)
    # ====== prepare the deconvolution ====== #
    # theano require batch_dims is Constant or None, but tensorflow
    # require batch_dims is a native TensorVariable
    output_shape = self.output_shape
    _ = tuple(output_shape)
    native_shape = tf.shape(X)
    output_shape = [native_shape[i] if j is None else j
                    for i, j in enumerate(output_shape)]
    deconved = deconv_func(value=X, filter=self.get('W'),
        output_shape=output_shape, strides=(1,) + self.strides + (1,),
        padding=self.pad)
    return K.set_shape(deconved, _)
